# Copyright (C) 2014-2020 The BET Development Team

r"""
This module contains tests for :module:`bet.calculateP.calculateP`.


Most of these tests should make sure certain values are within a tolerance
rather than exact due to the stocastic nature of the algorithms being tested.
"""
import os
import unittest
import bet
import bet.calculateP.calculateP as calcP
import bet.calculateP.simpleFunP as simpleFunP
import bet.sampling.basicSampling as bsam
import bet.sample as samp
import numpy as np
import numpy.testing as nptest
import bet.util as util
from bet.Comm import comm

#data_path = os.path.dirname(bet.__file__) + "/../test/test_calculateP/datafiles"
data_path = "test/test_calculateP/datafiles"


class prob:
    def test_prob_sum_to_1(self):
        """
        Test to see if the prob. sums to 1.
        """
        nptest.assert_almost_equal(np.sum(self.inputs._probabilities), 1.0)
    # @unittest.skipIf(comm.size > 1, 'Only run in serial')

    def test_vol_sum_to_1(self):
        """
        Test that volume ratios sum to 1.
        """
        nptest.assert_almost_equal(np.sum(self.inputs._volumes), 1.0)

    def test_prob_pos(self):
        """
        Test that all probs are non-negative.
        """
        self.assertEqual(np.sum(np.less(self.inputs._probabilities, 0)), 0)


class prob_on_emulated_samples:
    def test_P_sum_to_1(self):
        """
        Test that prob. sums to 1.
        """
        self.inputs_emulated.local_to_global()
        nptest.assert_almost_equal(
            np.sum(self.inputs_emulated._probabilities), 1.0)


    def test_prob_pos(self):
        """
        Test that all probabilites are non-negative.
        """
        self.inputs_emulated.local_to_global()
        self.assertEqual(
            np.sum(np.less(self.inputs_emulated._probabilities, 0)), 0)


class prob_with_emulated_volumes:
    def test_P_sum_to_1(self):
        """
        Test that probs sum to 1.
        """
        nptest.assert_almost_equal(np.sum(self.inputs._probabilities), 1.0)

    def test_vol_sum_to_1(self):
        """
        Test that volume ratios sum to 1.
        """
        nptest.assert_almost_equal(np.sum(self.inputs._volumes), 1.0)

    def test_prob_pos(self):
        """
        Test that all probs are non-negative.
        """
        self.assertEqual(np.sum(np.less(self.inputs._probabilities, 0)), 0)


class TestProbMethod_3to2(unittest.TestCase):
    """
    Sets up 3 to 2 map problem.
    """

    def setUp(self):
        self.inputs = samp.sample_set(3)
        self.outputs = samp.sample_set(2)
        self.inputs.set_values(np.loadtxt(data_path + "/3to2_samples.txt.gz"))
        self.outputs.set_values(np.loadtxt(data_path + "/3to2_data.txt.gz"))
        Q_ref = np.array([0.422, 0.9385])
        self.output_prob = simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            self.outputs, Q_ref=Q_ref, rect_scale=0.2, cells_per_dimension=1)

        self.inputs.set_domain(np.array([[0.0, 1.0],
                                         [0.0, 1.0],
                                         [0.0, 1.0]]))
        import numpy.random as rnd
        rnd.seed(1)
        self.inputs_emulated = bsam.random_sample_set('uniform', self.inputs.get_dim(),
                                                      num_samples=1001, globalize=True)
        self.disc = samp.discretization(input_sample_set=self.inputs,
                                        output_sample_set=self.outputs,
                                        output_probability_set=self.output_prob,
                                        emulated_input_sample_set=self.inputs_emulated)


class Test_prob_3to2(TestProbMethod_3to2, prob):
    """
    Test :meth:`bet.calculateP.calculateP.prob` on 3 to 2 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_3to2, self).setUp()
        self.disc._input_sample_set.estimate_volume_mc()
        calcP.prob(self.disc)


class Test_prob_on_emulated_samples_3to2(
        TestProbMethod_3to2, prob_on_emulated_samples):
    """
    Test :meth:`bet.calculateP.calculateP.prob_on_emulated_samples` on a 3 to 2 map.
    """

    def setUp(self):
        """
        Set up 3 to 2 map.
        """
        super(Test_prob_on_emulated_samples_3to2, self).setUp()
        calcP.prob_on_emulated_samples(self.disc)


class Test_prob_with_emulated_volumes_3to2(
        TestProbMethod_3to2, prob_with_emulated_volumes):
    """
    Test :meth:`bet.calculateP.calculateP.prob_with_emulated_volumes` on a 3 to 2 map.
    """

    def setUp(self):
        """
        Set up 3 to 2 problem.
        """
        super(Test_prob_with_emulated_volumes_3to2, self).setUp()
        calcP.prob_with_emulated_volumes(self.disc)


class TestProbMethod_3to1(unittest.TestCase):
    """
    Sets up 3 to 1 map problem.
    """

    def setUp(self):
        self.inputs = samp.sample_set(3)
        self.outputs = samp.sample_set(1)
        self.inputs.set_values(np.loadtxt(data_path + "/3to2_samples.txt.gz"))
        self.outputs.set_values(np.loadtxt(
            data_path + "/3to2_data.txt.gz")[:, 0])
        Q_ref = np.array([0.422])
        self.output_prob = simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            self.outputs, Q_ref=Q_ref, rect_scale=0.2, cells_per_dimension=1)

        self.inputs.set_domain(np.array([[0.0, 1.0],
                                         [0.0, 1.0],
                                         [0.0, 1.0]]))
        import numpy.random as rnd
        rnd.seed(1)
        self.inputs_emulated = bsam.random_sample_set('uniform',
                                                      self.inputs.get_dim(), num_samples=1001, globalize=True)
        self.disc = samp.discretization(input_sample_set=self.inputs,
                                        output_sample_set=self.outputs,
                                        output_probability_set=self.output_prob,
                                        emulated_input_sample_set=self.inputs_emulated)


class Test_prob_3to1(TestProbMethod_3to1, prob):
    """
    Test :meth:`bet.calculateP.calculateP.prob` on a 3 to 1 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_3to1, self).setUp()
        self.disc._input_sample_set.estimate_volume_mc()
        calcP.prob(self.disc)


class Test_prob_on_emulated_samples_3to1(
        TestProbMethod_3to1, prob_on_emulated_samples):
    """
    Test :meth:`bet.calculateP.calculateP.prob_on_emulated_samples` on a 3 to 1 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_on_emulated_samples_3to1, self).setUp()
        calcP.prob_on_emulated_samples(self.disc)


class Test_prob_with_emulated_volumes_3to1(
        TestProbMethod_3to1, prob_with_emulated_volumes):
    """
    Test :meth:`bet.calculateP.calculateP.prob_with_emulated_volumes` on a 3 to 1 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_with_emulated_volumes_3to1, self).setUp()
        calcP.prob_with_emulated_volumes(self.disc)


class TestProbMethod_10to4(unittest.TestCase):
    """
    Sets up 10 to 4 map problem.
    """

    def setUp(self):
        """
        Set up problem.
        """
        import numpy.random as rnd
        rnd.seed(1)
        self.inputs = samp.sample_set(10)
        self.outputs = samp.sample_set(4)
        self.lam_domain = np.zeros((10, 2))
        self.lam_domain[:, 0] = 0.0
        self.lam_domain[:, 1] = 1.0
        self.inputs.set_domain(self.lam_domain)
        num_samples = int(200 * comm.size)
        self.inputs = bsam.random_sample_set('uniform',
                                             self.inputs.get_dim(),
                                             num_samples=num_samples,
                                             globalize=True)
        self.outputs.set_values(np.dot(self.inputs._values, rnd.rand(10, 4)))
        Q_ref = np.mean(self.outputs._values, axis=0)
        self.inputs_emulated = bsam.random_sample_set('uniform',
                                                      self.inputs.get_dim(), num_samples=1001, globalize=True)
        self.output_prob = simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            self.outputs, Q_ref=Q_ref, rect_scale=0.2, cells_per_dimension=1)
        self.disc = samp.discretization(input_sample_set=self.inputs,
                                        output_sample_set=self.outputs,
                                        output_probability_set=self.output_prob,
                                        emulated_input_sample_set=self.inputs_emulated)



class Test_prob_10to4(TestProbMethod_10to4, prob):
    """
    Test :meth:`bet.calculateP.calculateP.prob` on a 10 to 4 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_10to4, self).setUp()
        self.disc._input_sample_set.estimate_volume_mc()
        calcP.prob(self.disc)


class Test_prob_on_emulated_samples_10to4(
        TestProbMethod_10to4, prob_on_emulated_samples):
    """
    Test :meth:`bet.calculateP.calculateP.prob_on_emulated_samples` on a 10 to 4 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_on_emulated_samples_10to4, self).setUp()

        calcP.prob_on_emulated_samples(self.disc)


class Test_prob_with_emulated_volumes_10to4(
        TestProbMethod_10to4, prob_with_emulated_volumes):
    """
    Test :meth:`bet.calculateP.calculateP.prob_with_emulated_volumes` on a 10 to 4 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_with_emulated_volumes_10to4, self).setUp()
        calcP.prob_with_emulated_volumes(self.disc)


class TestProbMethod_1to1(unittest.TestCase):
    """
    Sets up 1 to 1 map problem. Uses vectors instead of 2D arrays.
    """

    def setUp(self):
        """
        Set up problem.
        """

        import numpy.random as rnd
        rnd.seed(1)
        self.inputs = samp.sample_set(1)
        self.outputs = samp.sample_set(1)
        self.lam_domain = np.zeros((1, 2))
        self.lam_domain[:, 0] = 0.0
        self.lam_domain[:, 1] = 1.0
        self.inputs.set_domain(self.lam_domain)
        self.inputs.set_values(rnd.rand(100,))
        self.num_l_emulate = 1001
        self.inputs = bsam.random_sample_set('uniform',
                                             self.inputs.get_dim(), num_samples=1001, globalize=True)
        self.outputs.set_values(2.0 * self.inputs._values)
        Q_ref = np.mean(self.outputs._values, axis=0)
        self.inputs_emulated = bsam.random_sample_set('uniform',
                                                      self.inputs.get_dim(), num_samples=self.num_l_emulate,
                                                      globalize=True)
        self.output_prob = simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
            self.outputs, Q_ref=Q_ref, rect_scale=0.2, cells_per_dimension=1)
        self.disc = samp.discretization(input_sample_set=self.inputs,
                                        output_sample_set=self.outputs,
                                        output_probability_set=self.output_prob,
                                        emulated_input_sample_set=self.inputs_emulated)



class Test_prob_1to1(TestProbMethod_1to1, prob):
    """
    Test :meth:`bet.calculateP.calculateP.prob` on a 1 to 1 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_1to1, self).setUp()
        self.disc._input_sample_set.estimate_volume_mc()
        calcP.prob(self.disc)


class Test_prob_on_emulated_samples_1to1(
        TestProbMethod_1to1, prob_on_emulated_samples):
    """
    Test :meth:`bet.calculateP.calculateP.prob_on_emulated_samples` on a 1 to 1 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_on_emulated_samples_1to1, self).setUp()
        calcP.prob_on_emulated_samples(self.disc)


class Test_prob_with_emulated_volumes_1to1(
        TestProbMethod_1to1, prob_with_emulated_volumes):
    """
    Test :meth:`bet.calculateP.calculateP.prob_with_emulated_volumes` on a 1 to 1 map.
    """

    def setUp(self):
        """
        Set up problem.
        """
        super(Test_prob_with_emulated_volumes_1to1, self).setUp()
        calcP.prob_with_emulated_volumes(self.disc)


class Test_prob_from_sample_set(unittest.TestCase):
    """
    Test: method: `bet.calculateP.prob_from_sample_set`,
    : method: `bet.calculateP.prob_from_sample_set_with_emulated_volumes`,
    and : method: `bet.calculateP.prob_from_discretization_input`
    on a 2D domain.
    """

    def setUp(self):
        self.set_new = samp.rectangle_sample_set(dim=2)
        self.set_new.set_domain(np.array([[0.0, 1.0], [0.0, 1.0]]))
        self.set_new.setup(maxes=[[0.75, 0.75]], mins=[[0.25, 0.25]])

        self.set_old = samp.cartesian_sample_set(dim=2)
        self.set_old.set_domain(np.array([[0.0, 1.0], [0.0, 1.0]]))
        self.set_old.setup([np.linspace(0, 1, 21), np.linspace(0, 1, 21)])
        num_old = self.set_old.check_num()
        probs = np.zeros((num_old,))
        probs[0:-1] = 1.0 / float(num_old - 1)
        self.set_old.set_probabilities(probs)

        self.set_em = samp.cartesian_sample_set(dim=2)
        self.set_em.set_domain(np.array([[0.0, 1.0], [0.0, 1.0]]))
        self.set_em.setup([np.linspace(0, 1, 101), np.linspace(0, 1, 101)])

    def test_methods(self):
        calcP.prob_from_sample_set_with_emulated_volumes(
            self.set_old, self.set_new, self.set_em)
        nptest.assert_almost_equal(self.set_new._probabilities, [0.25, 0.75])
        calcP.prob_from_sample_set(self.set_old, self.set_new)
        nptest.assert_almost_equal(self.set_new._probabilities, [0.25, 0.75])
        disc = samp.discretization(input_sample_set=self.set_old,
                                   output_sample_set=self.set_old)
        calcP.prob_from_discretization_input(disc, self.set_new)
        nptest.assert_almost_equal(self.set_new._probabilities, [0.25, 0.75])
        num_em = self.set_em.check_num()
        probs = np.zeros((num_em,))
        probs[0:-1] = 1.0 / float(num_em - 1)
        self.set_em.set_probabilities(probs)
        self.set_em.global_to_local()
        disc = samp.discretization(input_sample_set=self.set_old,
                                   output_sample_set=self.set_old,
                                   emulated_input_sample_set=self.set_em)
        calcP.prob_from_discretization_input(disc, self.set_new)
        nptest.assert_almost_equal(self.set_new._probabilities, [0.25, 0.75])



# Copyright (C) 2014-2020 The BET Development Team

# -*- coding: utf-8 -*-
import numpy as np
from dolfin import *
from meshDS import meshDS
from projectKL import projectKL
from poissonRandField import solvePoissonRandomField
import scipy.io as sio
import sys


def computeSaveKL(numKL):
    '''
    ++++++++++++++++ Steps in Computing the Numerical KL Expansion ++++++++++
    We proceed by loading the mesh and defining the function space for which
    the eigenfunctions are defined upon.

    Then, we define the covariance kernel which requires correlation lengths
    and a standard deviation.

    We then compute and save the terms in a truncated KL expansion.
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    '''

    # Step 1: Set up the Mesh and Function Space

    mesh = Mesh("Lshaped.xml")

    # initialize the mesh to generate connectivity
    mesh.init()

    # Random field is projected on the space of Hat functions in the mesh
    V = FunctionSpace(mesh, "Lagrange", 1)

    # Step 2: Project covariance in the mesh and get the eigenfunctions

    # Initialize the projectKL object with the mesh
    Lmesh = projectKL(mesh)

    # Create the covariance expression to project on the mesh.
    etaX = 10.0
    etaY = 10.0
    C = 1

    # Pick your favorite covariance. Popular choices are Gaussian (of course),
    # Exponential, triangular (has finite support which is nice). Check out
    # Ghanem and Spanos' book for more classical options.

    # A Gaussian Covariance
    '''
    cov = Expression("C*exp(-((x[0]-x[1]))*((x[0]-x[1]))/ex - \
      ((x[2]-x[3]))*((x[2]-x[3]))/ey)",
             ex=etaX,ey=etaY, C=C)
    '''
    # An Exponential Covariance
    cov = Expression("C*exp(-fabs(x[0]-x[1])/ex - fabs(x[2]-x[3])/ey)", ex=etaX, ey=etaY, C=C,
                     degree=10)

    # Solve the discrete covariance relation on the mesh
    Lmesh.projectCovToMesh(numKL, cov)

    # Get the eigenfunctions and eigenvalues
    eigen_val = Lmesh.eigen_vals

    eigen_func_mat = np.zeros(
        (numKL, Lmesh.eigen_funcs[0].vector().array().size))
    for i in range(0, numKL):
        eigen_func_mat[i, :] = Lmesh.eigen_funcs[i].vector().array()

    kl_mdat = dict()
    kl_mdat['KL_eigen_funcs'] = eigen_func_mat
    kl_mdat['KL_eigen_vals'] = eigen_val

    sio.savemat("KL_expansion", kl_mdat)



# Copyright (C) 2014-2020 The BET Development Team

"""
Butler, Estep, Tavener Method

This package provides tools for solving stochastic inverse problems using a
measure-theoretic. It is named for the developers of the key algorithm in
:mod:`bet.calculateP.calculateP`.

* :mod:`~bet.Comm` provides a work around for users who do not which to install :program:``mpi4py``.
* :mod:`~bet.util` provides some general use methods for creating grids, checking/fixing dimensions, and globalizing arrays.
* :mod:`~bet.calculateP` provides tools to approximate probabilities.
* :mod:`~bet.sampling` provides various sampling algorithms.
* :mod:`~bet.sensitivity` provides tools for approximating derivatives and optimally choosing quantities of interest.
* :mod:`~bet.postProcess` provides plotting tools and tools to sort samples by probabilities.
* :mod:`~bet.sample` provides data structures to store sets of samples and their associated arrays.
* :mod:`~bet.surrogates` provides methods for generating and using
    surrogate models.

"""

__all__ = ['sampling', 'calculateP', 'postProcess', 'sensitivity', 'util',
           'Comm', 'sample', 'surrogates']



# Copyright (C) 2014-2020 The BET Development Team

r"""

Evaluate the model at input values.

"""

import os
import scipy.io as sio
import sys
import numpy as np
import numpy.random as rand

"""
Defines an exact and approximate nonlinear map.

"""


def lb_model_exact(input_data):
    """
    Exact map.
    """
    num_runs = input_data.shape[0]
    num_runs_dim = input_data.shape[1]

    values = np.zeros((num_runs, 2))
    jacobians = np.zeros((num_runs, 2, 3))
    for i in range(num_runs):
        x = input_data[i, 0]
        y = input_data[i, 1]
        z = input_data[i, 2]
        values[i, 0] = 2.0 * x * x + y * y + z * z
        values[i, 1] = x * y + y + 2.0 * z
        jac = np.array([[2.0 * x, y, z],
                        [y, 1.0, 2.0]])
        jacobians[i] = jac
    return (values, jacobians)


def lb_model(input_data):
    """

    Approximate map. Adds error to the exact map.

    """
    (values, jacobians) = lb_model_exact(input_data)
    num_runs = input_data.shape[0]
    error_estimates = 0.2 * rand.randn(num_runs, 2)
    values = values - error_estimates
    return(values, error_estimates, jacobians)



#! /usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

r"""
This example evaluates a nonlinear map to a 1d or 2d space.
Modify QoI_num in myModel.py to change the dimension of the output.
The maps are defined
as quantities of interest (QoI) defined as spatial
observations of the solution to the elliptic PDE .. math::
  :nowrap:

  \begin{cases}
    -\nabla \cdot (A(\lambda)\cdot\nabla u) &= f(x,y;\lambda), \ (x,y)\in\Omega, \\
    u|_{\partial \Omega} &= 0,
  \end{cases}

where :math:`A(\lambda)=\text{diag}(1/\lambda_1^2,1/\lambda_2^2)`,
:math: `f(x,y;\lambda) = \pi^2 \sin(\pi x\lambda_1)\sin(\pi y \lambda_2)`,
and :math:`\Omega=[0,1]\times[0,1]`.

The 2-D input space is sampled with initial i.i.d. uniform
random samples.
We refer to the input space as the parameter space,
and use parameter to refer to a particular
point (e.g., a particular random sample) in this space.
The model is solved for these parameters to generate
"predicted" data.

The parameter space is also sampled with a different ("data-generating") random variable, and the linear map
is applied to generate artificial "observed" data.
We solve the density-based approach problem defined by the predicted inputs and outputs and the
observed output data.
In this problem, the initial uniform probability on the parameter space is updated to a new probability measure
based on the data-consistent inversion framework.
This can be compared to the data-generating distribution through plots and a variety of distance metrics.
"""

import numpy as np
import bet.postProcess.plotP as plotP
import bet.calculateP.calculateR as calculateR
import bet.sample as samp
import bet.sampling.basicSampling as bsam
import bet.postProcess.compareP as compP
from myModel import my_model

# Define the sampler that will be used to create the discretization
# object, which is the fundamental object used by BET to compute
# solutions to the stochastic inverse problem.
# The sampler and my_model is the interface of BET to the model,
# and it allows BET to create input/output samples of the model.
sampler = bsam.sampler(my_model)

# Initialize 2-dimensional input parameter sample set object
input_samples = samp.sample_set(2)

# Set parameter domain
input_samples.set_domain(np.array([[3.0, 6.0],
                                   [1.0, 5.0]]))

num_samples = int(1E3)
'''
Suggested changes for user:

Try num_samples = 1E3 and 1E4.
What happens when num_samples = 1E2?
'''

# Generate samples on the parameter space
input_samples = sampler.random_sample_set('uniform', input_samples, num_samples=num_samples)

# Create the prediction discretization object using the input samples
disc_predict = sampler.compute_qoi_and_create_discretization(input_samples)

# Generate observed data
sampler_obs = bsam.sampler(my_model)

# Initialize 3-dimensional input parameter sample set object
input_samples_obs = samp.sample_set(2)

# Set parameter domain
input_samples_obs.set_domain(input_samples.get_domain())

# Generate samples on the parameter space
beta_a = 2.0  # a parameter for beta distribution
beta_b = 2.0  # b parameter for beta distribution

'''
Suggested changes for user:

Try changing beta_a and beta_b to shift the data-generating distribution.
Try beta_a = 5, beta_b = 2.
Try beta_a = 0.5, beta_b = 3

Both parameters must be non-negative.
'''
domain_obs = input_samples_obs.get_domain()

input_samples_obs = sampler_obs.random_sample_set([['beta', {'a': beta_a, 'b': beta_b, 'loc': domain_obs[0, 0],
                                                             'scale': domain_obs[0, 1] - domain_obs[0, 0]}],
                                                   ['beta', {'a': beta_a, 'b': beta_b, 'loc': domain_obs[1, 0],
                                                             'scale': domain_obs[1, 1] - domain_obs[1, 0]}]],
                                                  input_samples_obs, num_samples=int(1E3))
disc_obs = sampler_obs.compute_qoi_and_create_discretization(input_samples_obs)

# Set probability set for predictions
disc_predict.set_output_observed_set(disc_obs.get_output_sample_set())


# Calculate initial total variation of marginals
comp_init = compP.compare(disc_predict, disc_obs, set1_init=True, set2_init=True)
print("Initial TV of Marginals")
for i in range(2):
    print(comp_init.distance_marginal_quad(i=i, compare_factor=0.2, rtol=1.0e-3, maxiter=100))

print("------------------------------------------------------")

invert_to = 'kde'  # 'multivariate_gaussian', 'expon', 'beta'

'''
Suggested changes for user:

Try changing the type of probability measure to invert to from
'kde': Gaussian kernel density estimate (generally the best and most robust choice)
'multivariate_gaussian': fit a multivariate Gaussian distribution
'beta': fit a beta distribution
'expon': fit an exponential distribution (useful if beta_a or beta_b <= 1)

'''

if invert_to == 'kde':
    # Invert to weighted KDE
    print("Weighted Kernel Density Estimate")
    calculateR.invert_to_kde(disc_predict)
elif invert_to == 'multivariate_gaussian':
    # Invert to multivariate Gaussian
    print("Multivariate Gaussian")
    calculateR.invert_to_multivariate_gaussian(disc_predict)
elif invert_to == 'beta':
    # Invert and fit Beta distribution
    print("Beta Distribution")
    calculateR.invert_to_random_variable(disc_predict, rv='beta')
elif invert_to == 'expon':
    # Invert and fit Beta distribution
    print("Beta Distribution")
    calculateR.invert_to_random_variable(disc_predict, rv='expon')
else:
    raise RuntimeError("Not an acceptable type of Inversion.")


# Calculate Total Variation between updated marginals and data-generating marginals
for i in range(2):
    plotP.plot_1d_marginal_densities(sets=(disc_predict.get_input_sample_set(),
                                           disc_obs.get_input_sample_set()),
                                     i=i,
                                     sets_label_initial=['Initial', 'Data-Generating'],
                                     sets_label=['Updated', ''])
# Calculate updated total variation
comp_init = compP.compare(disc_predict, disc_obs, set1_init=False, set2_init=True)
print("Updated TV of Marginals")
for i in range(2):
    print(comp_init.distance_marginal_quad(i=i, compare_factor=0.2, rtol=1.0e-3, maxiter=1000))



# Copyright (C) 2014-2020 The BET Development Team

r"""
This module provides methods for calculating the probability measure
:math:`P_{\Lambda}`.

* :mod:`~bet.calculateP.prob_on_emulated_samples` provides a skeleton class and calculates the probability for a set of emulation points.
* :mod:`~bet.calculateP.calculateP.prob` estimates the probability based on pre-defined volumes.
* :mod:`~bet.calculateP.calculateP.prob_with_emulated` estimates the probability using volume emulation.
* :mod:`~bet.calculateP.calculateP.prob_from_sample_set` estimates the probability based on probabilities from another
sample set on the same space.

"""
import logging
import numpy as np
from bet.Comm import comm, MPI
import bet.sample as samp
import bet.util as util


def prob_on_emulated_samples(discretization, globalize=True):
    r"""

    Calculates :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{emulate}})`, the
    probability associated with a set of Voronoi cells defined by
    ``num_l_emulate`` iid samples :math:`(\lambda_{emulate})`.
    This is added to the emulated input sample set object.

    :param discretization: An object containing the discretization information.
    :type discretization: class:`bet.sample.discretization`
    :param bool globalize: Makes local variables global.

    """

    # Check dimensions
    discretization.check_nums()
    op_num = discretization._output_probability_set.check_num()
    discretization._emulated_input_sample_set.check_num()

    # Check for necessary properties
    if discretization._io_ptr_local is None:
        discretization.set_io_ptr(globalize=True)
    if discretization._emulated_ii_ptr_local is None:
        discretization.set_emulated_ii_ptr(globalize=False)

    # Calculate Probabilties
    P = np.zeros((discretization._emulated_input_sample_set.
                  _values_local.shape[0],))
    d_distr_emu_ptr = discretization._io_ptr[discretization.
                                             _emulated_ii_ptr_local]
    for i in range(op_num):
        if discretization._output_probability_set._probabilities[i] > 0.0:
            Itemp = np.equal(d_distr_emu_ptr, i)
            Itemp_sum = np.sum(Itemp)
            Itemp_sum = comm.allreduce(Itemp_sum, op=MPI.SUM)
            if Itemp_sum > 0:
                P[Itemp] = discretization._output_probability_set.\
                    _probabilities[i] / Itemp_sum

    discretization._emulated_input_sample_set._probabilities_local = P
    discretization._emulated_input_sample_set.set_prob_type('voronoi')
    if globalize:
        discretization._emulated_input_sample_set.local_to_global()
    pass


def prob(discretization, globalize=True):
    r"""
    Calculates :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples}})`, the
    probability associated with a set of  cells defined by the model
    solves at :math:`(\lambda_{samples})` where the volumes of these
    cells are provided.

    :param discretization: An object containing the discretization information.
    :type discretization: class:`bet.sample.discretization`
    :param bool globalize: Makes local variables global.

    """

    # Check Dimensions
    discretization.check_nums()
    op_num = discretization._output_probability_set.check_num()

    # Check for necessary attributes
    if discretization._io_ptr_local is None:
        discretization.set_io_ptr(globalize=False)

    # Calculate Probabilities
    if discretization._input_sample_set._values_local is None:
        discretization._input_sample_set.global_to_local()
    P_local = np.zeros((len(discretization._io_ptr_local),))
    for i in range(op_num):
        if discretization._output_probability_set._probabilities[i] > 0.0:
            Itemp = np.equal(discretization._io_ptr_local, i)
            Itemp_sum = np.sum(discretization._input_sample_set.
                               _volumes_local[Itemp])
            Itemp_sum = comm.allreduce(Itemp_sum, op=MPI.SUM)
            if Itemp_sum > 0:
                P_local[Itemp] = discretization._output_probability_set.\
                    _probabilities[i] * discretization._input_sample_set.\
                    _volumes_local[Itemp] / Itemp_sum
    if globalize:
        discretization._input_sample_set._probabilities = util.\
            get_global_values(P_local)
    discretization._input_sample_set._probabilities_local = P_local
    discretization._input_sample_set.set_prob_type('voronoi')



def prob_with_emulated_volumes(discretization):
    r"""

    Calculates :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples}})`, the
    probability associated with a set of  cells defined by the model
    solves at :math:`(\lambda_{samples})` where the volumes are calculated
    with the given emulated input points.

    :param discretization: An object containing the discretization information.
    :type discretization: class:`bet.sample.discretization`
    :param globalize: Makes local variables global.

    """

    # Check Dimensions
    discretization.check_nums()
    discretization._output_probability_set.check_num()
    if discretization._output_probability_set._values_local is None:
        discretization._output_probability_set.global_to_local()

    # Calculate Volumes
    discretization.estimate_input_volume_emulated()
    return prob(discretization)


def prob_from_sample_set_with_emulated_volumes(set_old, set_new,
                                               set_emulate=None):
    r"""

    Calculates :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples_new}})`
    from :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples_old}})` using
    a set of emulated points are distributed with respect to the
    volume measure.

    :param set_old: Sample set on which probabilities have already been
        calculated
    :type set_old: :class:`~bet.sample.sample_set_base`
    :param set_new: Sample set for which probabilities will be calculated.
    :type set_new: :class:`~bet.sample.sample_set_base`
    :param set_emulate: Sample set for volume emulation
    :type set_emulate: :class:`~bet.sample.sample_set_base`

    """
    if set_emulate is None:
        logging.warning("Using MC assumption because no emulated points given")
        return prob_from_sample_set(set_old, set_new)

    # Check dimensions
    num_old = set_old.check_num()
    num_new = set_new.check_num()
    set_emulate.check_num()
    if (set_old._dim != set_new._dim) or (set_old._dim != set_emulate._dim):
        raise samp.dim_not_matching("Dimensions of sets are not equal.")
    # Localize emulated points
    if set_emulate._values_local is None:
        set_emulate.global_to_local()

    # Map emulated points to old and new sets
    (_, ptr1) = set_old.query(set_emulate._values_local)
    (_, ptr2) = set_new.query(set_emulate._values_local)
    ptr1 = ptr1.flat[:]
    ptr2 = ptr2.flat[:]

    # Set up probability vectors
    prob_new = np.zeros((num_new,))
    prob_em = np.zeros((len(ptr1), ))

    # Loop over old cells and divide probability over emulated cells
    warn = False
    for i in range(num_old):
        if set_old._probabilities[i] > 0.0:
            Itemp = np.equal(ptr1, i)
            Itemp_sum = np.sum(Itemp)
            Itemp_sum = comm.allreduce(Itemp_sum, op=MPI.SUM)
            if Itemp_sum > 0:
                prob_em[Itemp] += set_old._probabilities[i] / float(Itemp_sum)
            else:
                warn = True
    # Warn that some cells have no emulated points in them
    if warn:
        msg = "Some old cells have no emulated points in them. "
        msg += "Renormalizing probability."
        logging.warning(msg)
        total_prob = np.sum(prob_em)
        total_prob = comm.allreduce(total_prob, op=MPI.SUM)
        prob_em = prob_em / total_prob
    # Loop over new cells and distribute probability from emulated cells
    for i in range(num_new):
        Itemp = np.equal(ptr2, i)
        Itemp_sum = np.sum(prob_em[Itemp])
        Itemp_sum = comm.allreduce(Itemp_sum, op=MPI.SUM)
        prob_new[i] = Itemp_sum

    # Set probabilities
    set_new.set_probabilities(prob_new)
    set_new.set_prob_type('voronoi')
    return prob_new


def prob_from_sample_set(set_old, set_new):
    r"""

    Calculates :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples_new}})`
    from :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples_old}})` using
    the MC assumption with respect to set_old.

    :param set_old: Sample set on which probabilities have already been
        calculated
    :type set_old: :class:`~bet.sample.sample_set_base`
    :param set_new: Sample set for which probabilities will be calculated.
    :type set_new: :class:`~bet.sample.sample_set_base`

    """
    # Check dimensions
    set_old.check_num()
    num_new = set_new.check_num()

    if (set_old._dim != set_new._dim):
        raise samp.dim_not_matching("Dimensions of sets are not equal.")

    # Map old points new sets
    if set_old._values_local is None:
        set_old.global_to_local()
    (_, ptr) = set_new.query(set_old._values_local)
    ptr = ptr.flat[:]

    # Set up probability vector
    prob_new = np.zeros((num_new,))

    # Loop over new cells and distribute probability from old
    for i in range(num_new):
        Itemp = np.equal(ptr, i)
        Itemp_sum = np.sum(set_old._probabilities_local[Itemp])
        Itemp_sum = comm.allreduce(Itemp_sum, op=MPI.SUM)
        prob_new[i] = Itemp_sum

    # Set probabilities
    set_new.set_probabilities(prob_new)
    set_new.set_prob_type('voronoi')
    return prob_new


def prob_from_discretization_input(disc, set_new):
    r"""

    Calculates :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples_new}})`
    from :math:`P_{\Lambda}(\mathcal{V}_{\lambda_{samples_old}})` where
    :math:`\lambda_{samples_old}` come from an input discretization.

    :param disc: Discretiztion on which probabilities have already been
        calculated
    :type disc: :class:`~bet.sample.discretization`
    :param set_new: Sample set for which probabilities will be calculated.
    :type set_new: :class:`~bet.sample.sample_set_base`

    """
    if disc._emulated_input_sample_set is None:
        logging.warning("Using MC assumption because no emulated points given")
        em_set = disc._input_sample_set
    else:
        em_set = disc._emulated_input_sample_set

    if em_set._values_local is None:
        em_set.global_to_local()
    if em_set._probabilities_local is None:
        raise AttributeError("Probabilities must be pre-calculated.")

    # Check dimensions
    disc.check_nums()
    num_new = set_new.check_num()

    if (disc._input_sample_set._dim != set_new._dim):
        raise samp.dim_not_matching("Dimensions of sets are not equal.")

    (_, ptr) = set_new.query(em_set._values_local)
    ptr = ptr.flat[:]

    # Set up probability vectors
    prob_new = np.zeros((num_new,))
    prob_em = em_set._probabilities_local

    for i in range(num_new):
        Itemp = np.equal(ptr, i)
        Itemp_sum = np.sum(prob_em[Itemp])
        Itemp_sum = comm.allreduce(Itemp_sum, op=MPI.SUM)
        prob_new[i] = Itemp_sum

    # Set probabilities
    set_new.set_probabilities(prob_new)
    set_new.set_prob_type('voronoi')
    return prob_new



#! /usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

"""
This example takes uniformly distributed samples of parameters and
output data from a simple groundwater contaminant transport model,
and calculates solutions to the stochastic inverse problem.
The parameter domain is 5D, where the uncertain parameters are the x and y
locations of the source, the horizontal dispersivity, the flow angle,
and the contaminant flux. There are 11 measured QoIs in the data space
available. By changing the choice of QoIs that we use to solve the stochastic
inverse problem, we see the effect of geometric distinctness.
Probabilities in the parameter space are
calculated using the MC assumption.  1D and 2D marginals are calculated,
smoothed, and plotted. The samples are then ranked by probability density
and the volume of high-probability regions are calculated. Probabilistic predictions of other QoI are made.

"""

import numpy as np
import bet.calculateP as calculateP
import bet.postProcess as postProcess
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateP as calculateP
import bet.postProcess.plotP as plotP
import bet.postProcess.plotDomains as plotD
import bet.postProcess.postTools as postTools
import bet.sample as samp

# Labels and descriptions of the uncertain parameters
labels = [
    'Source $y$ coordinate [L]',
    'Source $x$ coordinate [L]',
    'Dispersivity x [L]',
    'Flow Angle [degrees]',
    'Contaminant flux [M/T]']

# Load data from files
# First obtain info on the parameter domain
parameter_domain = np.loadtxt("files/lam_domain.txt.gz")  # parameter domain
parameter_dim = parameter_domain.shape[0]
# Create input sample set
input_samples = samp.sample_set(parameter_dim)
input_samples.set_domain(parameter_domain)
input_samples.set_values(np.loadtxt("files/samples.txt.gz"))
input_samples.estimate_volume_mc()  # Use standard MC estimate of volumes
# Choose which QoI to use and create output sample set
QoI_indices_observe = np.array([0, 1, 2, 3])
output_samples = samp.sample_set(QoI_indices_observe.size)
output_samples.set_values(np.loadtxt(
    "files/data.txt.gz")[:, QoI_indices_observe])

# Create discretization object
my_discretization = samp.discretization(input_sample_set=input_samples,
                                        output_sample_set=output_samples)

# Load the reference parameter and QoI values
param_ref = np.loadtxt("files/lam_ref.txt.gz")  # reference parameter set
# reference QoI set
Q_ref = np.loadtxt("files/Q_ref.txt.gz")[QoI_indices_observe]

# Plot the data domain
plotD.scatter_rhoD(
    my_discretization,
    ref_sample=Q_ref,
    io_flag='output',
    showdim=2)

# Whether or not to use deterministic description of simple function approximation of
# ouput probability
deterministic_discretize_D = True
if deterministic_discretize_D == True:
    simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(data_set=my_discretization,
                                                                       Q_ref=Q_ref,
                                                                       rect_scale=0.25,
                                                                       cells_per_dimension=1)
else:
    simpleFunP.uniform_partition_uniform_distribution_rectangle_scaled(data_set=my_discretization,
                                                                       Q_ref=Q_ref,
                                                                       rect_scale=0.25,
                                                                       M=50,
                                                                       num_d_emulate=1E5)

# calculate probabilities making Monte Carlo assumption
calculateP.prob(my_discretization)

# calculate 2D marginal probabilities
(bins, marginals2D) = plotP.calculate_2D_marginal_probs(my_discretization, nbins=10)

# smooth 2D marginal probabilites for plotting (optional)
marginals2D = plotP.smooth_marginals_2D(marginals2D, bins, sigma=1.0)

# plot 2D marginal probabilities
plotP.plot_2D_marginal_probs(marginals2D, bins, my_discretization, filename="contaminant_map",
                             plot_surface=False,
                             lam_ref=param_ref,
                             lambda_label=labels,
                             interactive=False)

# calculate 1d marginal probs
(bins, marginals1D) = plotP.calculate_1D_marginal_probs(my_discretization, nbins=20)

# smooth 1d marginal probs (optional)
marginals1D = plotP.smooth_marginals_1D(marginals1D, bins, sigma=1.0)

# plot 1d marginal probs
plotP.plot_1D_marginal_probs(marginals1D, bins, my_discretization,
                             filename="contaminant_map",
                             interactive=False,
                             lam_ref=param_ref,
                             lambda_label=labels)

percentile = 1.0
# Sort samples by highest probability density and sample highest
# percentile percent samples
(num_samples, my_discretization_highP, indices) = postTools.sample_highest_prob(
    percentile, my_discretization, sort=True)

# print the number of samples that make up the  highest percentile percent samples and
# ratio of the volume of the parameter domain they take up
print((num_samples, np.sum(my_discretization_highP._input_sample_set.get_volumes())))

# Choose unused QoI as prediction QoI and propagate measure onto predicted
# QoI data space
QoI_indices_predict = np.array([7])
output_samples_predict = samp.sample_set(QoI_indices_predict.size)
output_samples_predict.set_values(np.loadtxt(
    "files/data.txt.gz")[:, QoI_indices_predict])
output_samples_predict.set_probabilities(input_samples.get_probabilities())

# Determine range of predictions and store as domain for plotting purposes
output_samples_predict.set_domain(output_samples_predict.get_bounding_box())

# Plot 1D pdf of predicted QoI
# calculate 1d marginal probs
(bins_pred, marginals1D_pred) = plotP.calculate_1D_marginal_probs(output_samples_predict,
                                                                  nbins=20)

# plot 1d pdf
plotP.plot_1D_marginal_probs(marginals1D_pred, bins_pred, output_samples_predict,
                             filename="contaminant_prediction", interactive=False)



# Copyright (C) 2014-2020 The BET Development Team

import bet.postProcess.compareP as compP
from helpers import *

"""
The ``helpers.py`` file contains functions that define
sample sets of an arbitrary dimension with probabilities
uniformly distributed in a hypercube of sidelength ``delta``.
The hypercube can be in three locations:
- corner at [0, 0, ..., 0]  in ``unit_bottom_set``
- centered at [0.5, 0.5, ... 0.5] in ``unit_center_set``
- corner in [1, 1, ..., 1] in `` unit_top_set``

and the number of samples will determine the fidelity of the
approximation since we are using voronoi-cell approximations.
"""
num_samples1 = 50
num_samples2 = 50
delta1 = 0.5  # width of measure's support per dimension
delta2 = 0.45
dim = 2
# define two sets that will be compared
set1 = unit_center_set(dim, num_samples1, delta1)
set2 = unit_center_set(dim, num_samples2, delta2)

# choose a reference sigma-algebra to compare both solutions
# against (using nearest-neighbor query).
num_comparison_samples = 2000
# the compP.compare method instantiates the compP.comparison class.
mm = compP.compare(set1, set2)  # initialize metric

# Use existing common library functions

# Use a function of your own!

def inftynorm(x, y):
    """
    Infinity norm between two vectors.
    """
    return np.max(np.abs(x - y))


mm.set_compare_set(compare_set=num_comparison_samples, compare_factor=0.1)

print(mm.distance('tv'))
print(mm.distance(inftynorm, normalize=False))
print(mm.distance('mink', w=0.5, p=1))
print(mm.distance('hell'))





# Copyright (C) 2014-2020 The BET Development Team

r"""
This module contains functions for the density-based approach that utilizes a ratio of observed to predicted densities
to update an initial density on the parameter space.

* :meth:`~bet.calculateP.calculateR.generate_output_kdes` generates KDEs on output sets.
* :meth:`~bet.calculateP.calculateR.invert_to_kde` solves SIP for weighted KDEs.
* :meth:`~bet.calculateP.calculateR.invert_to_gmm` solves SIP for a Gaussian Mixture Model.
* :meth:`~bet.calculateP.calculateR.invert_to_multivariate_gaussian` solves SIP for a multivariate Gaussian.
* :meth:`~bet.calculateP.calculateR.invert_to_random_variable` solves SIP for random variables.
* :meth:`~bet.calculateP.calculateR.invert_rejection_sampling` solves SIP with rejection sampling.

"""
import bet.sample
import numpy as np
import logging


def generate_output_kdes(discretization, bw_method=None):
    """
    Generate Kernel Density Estimates on predicted and observed output sample sets.

    :param discretization: Discretization used to calculate KDes
    :type discretization: :class:`bet.sample.discretization`
    :param bw_method: bandwidth method for `scipy.stats.gaussian_kde`.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html.
    :type bw_method: str

    :returns: prediction set, prediction kdes, observation set, observation kdes, number of clusters
    :rtype: :class:`bet.discretization.sample_set`, list, :class:`bet.discretization.sample_set`, list, int
    """
    from scipy.stats import gaussian_kde
    discretization.local_to_global()

    predict_set = discretization.get_output_sample_set()
    obs_set = discretization.get_output_observed_set()
    if predict_set.get_region() is None or obs_set.get_region() is None:
        predict_set.set_region(np.array([0] * predict_set.check_num()))
        obs_set.set_region(np.array([0] * obs_set.check_num()))

    if predict_set.get_cluster_maps() is None:
        num_clusters = int(np.max(predict_set.get_region()) + 1)
    else:
        num_clusters = len(predict_set.get_cluster_maps())

    predict_kdes = []
    obs_kdes = []
    for i in range(num_clusters):
        if predict_set.get_cluster_maps() is not None:
            if len(predict_set.get_cluster_maps()) > 1:
                if predict_set.get_weights_init() is None:
                    predict_kdes.append(gaussian_kde(predict_set.get_cluster_maps()[i].T, bw_method=bw_method))
                else:
                    predict_pointer = np.where(predict_set.get_region() == i)[0]
                    weights = predict_set.get_weights_init()[predict_pointer]
                    predict_kdes.append(gaussian_kde(predict_set.get_cluster_maps()[i].T, bw_method=bw_method,
                                                     weights=weights))
            else:
                predict_kdes.append(None)
        else:
            predict_pointer = np.where(predict_set.get_region() == i)[0]
            if len(predict_pointer) > 1:
                if predict_set.get_weights_init() is None:
                    predict_kdes.append(gaussian_kde(predict_set.get_values()[predict_pointer].T, bw_method=bw_method))
                else:
                    weights = predict_set.get_weights_init()[predict_pointer]
                    predict_kdes.append(gaussian_kde(predict_set.get_values()[predict_pointer].T, bw_method=bw_method,
                                                     weights=weights))
            else:
                predict_kdes.append(None)

        if obs_set.get_cluster_maps() is not None:
            if len(obs_set.get_cluster_maps()) > 1:
                obs_kdes.append(gaussian_kde(obs_set.get_cluster_maps()[i].T, bw_method=bw_method))
            else:
                obs_kdes.append(None)
        else:
            obs_pointer = np.where(obs_set.get_region() == i)[0]
            if len(obs_pointer) > 1:
                obs_kdes.append(gaussian_kde(obs_set.get_values()[obs_pointer].T, bw_method=bw_method))
            else:
                obs_kdes.append(None)
    return predict_kdes, obs_kdes, num_clusters


def invert(discretization, bw_method=None):
    """
    Solve the data consistent stochastic inverse problem, solving for input sample weights.

    :param discretization: Discretization on which to perform inversion.
    :type discretization: :class:`bet.sample.discretization`
    :param bw_method: bandwidth method for `scipy.stats.gaussian_kde`.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html.
    :type bw_method: str

    :return: acceptance rate, mean acceptance rate, pointers for samples to clusters
    :rtype: list, `np.ndarray`, list
    """
    predict_kdes, obs_kdes, num_clusters = generate_output_kdes(discretization, bw_method)
    predict_set = discretization.get_output_sample_set()

    rs = []
    r = []
    lam_ptr = []
    weights = np.zeros((discretization.get_output_sample_set().check_num(), ))
    for i in range(num_clusters):
        predict_pointer = np.where(predict_set.get_region() == i)[0]
        # First compute the rejection ratio
        if predict_set.get_cluster_maps() is None:
            vals = predict_set.get_values()[predict_pointer]
        else:
            vals = predict_set.get_cluster_maps()[i]
        if len(predict_pointer) > 0:
            r.append(np.divide(obs_kdes[i](vals.T), predict_kdes[i](vals.T)))
            rs.append((r[i].mean()))
        else:
            r.append(None)
            rs.append(None)
        weights[predict_pointer] = r[i]
        lam_ptr.append(predict_pointer)
    discretization.get_input_sample_set().set_weights(weights)
    return rs, r, lam_ptr


def invert_to_kde(discretization, bw_method=None):
    """
    Solve the data consistent stochastic inverse problem, solving for a weighted kernel density estimate.

    :param discretization: Discretization on which to perform inversion.
    :type discretization: :class:`bet.sample.discretization`
    :param bw_method: bandwidth method for `scipy.stats.gaussian_kde`.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html.
    :type bw_method: str

    :return: marginal probabilities and cluster weights
    :rtype: list, `np.ndarray`
    """
    from scipy.stats import gaussian_kde

    predict_kdes, obs_kdes, num_clusters = generate_output_kdes(discretization, bw_method)

    rs, r, lam_ptr = invert(discretization, bw_method)

    obs_set = discretization.get_output_observed_set()

    # Compute marginal probabilities for each parameter and initial condition.
    param_marginals = []
    cluster_weights = []
    num_obs = obs_set.check_num()

    input_dim = discretization.get_input_sample_set().get_dim()
    params = discretization.get_input_sample_set().get_values()

    for i in range(num_clusters):
        cluster_weights.append(len(np.where(obs_set.get_region() == i)[0]) / num_obs)
    for i in range(input_dim):
        param_marginals.append([])
        for j in range(num_clusters):
            if r[j] is not None:
                param_marginals[i].append(gaussian_kde(params[lam_ptr[j], i], weights=r[j], bw_method=bw_method))
            else:
                param_marginals[i].append(None)
    discretization.get_input_sample_set().set_prob_type("kde")
    discretization.get_input_sample_set().set_prob_parameters((param_marginals, cluster_weights))
    print('Diagnostic for clusters [sample average of ratios in each cluster]: ', rs)
    return param_marginals, cluster_weights


def invert_rejection_sampling(discretization, bw_method=None):
    """
    Solve the data consistent stochastic inverse problem by rejection sampling.

    :param discretization: Discretization on which to perform inversion.
    :type discretization: :class:`bet.sample.discretization`
    :param bw_method: bandwidth method for `scipy.stats.gaussian_kde`.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html.
    :type bw_method: str

    :return: sample set containing samples
    :rtype: :class:`bet.sample.sample_set`
    """
    predict_kdes, obs_kdes, num_clusters = generate_output_kdes(discretization, bw_method)

    rs, r, lam_ptr = invert(discretization, bw_method)

    discretization.get_input_sample_set().local_to_global()
    new_vals = []
    for i in range(num_clusters):
        check = np.random.uniform(low=0, high=1, size=r[i].size)  # create random uniform weights to check r against
        new_r = r[i] / np.max(r[i])  # normalize weights
        idx = np.where(new_r >= check)[0]  # rejection criterion
        new_vals.append(discretization.get_input_sample_set().get_values()[lam_ptr[i][idx]])
    vals = np.vstack(new_vals)
    new_set = bet.sample.sample_set(discretization.get_input_sample_set().get_dim())
    new_set.set_values(vals)
    n = vals.shape[0]
    probs = np.ones((n, )) / float(n)
    new_set.set_probabilities(probs)
    domain = []
    for i in range(new_set.get_dim()):
        x_max = np.max(vals[:, i])
        x_min = np.min(vals[:, i])
        domain.append([x_min, x_max])
    domain = np.array(domain)
    new_set.set_domain(domain)
    new_set.global_to_local()

    return new_set


def invert_to_gmm(discretization, bw_method=None):
    """
    Solve the data consistent stochastic inverse problem, solving for a Gaussian mixture model.

    :param discretization: Discretization on which to perform inversion.
    :type discretization: :class:`bet.sample.discretization`
    :param bw_method: bandwidth method for `scipy.stats.gaussian_kde`.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html.
    :type bw_method: str

    :return: means, covariances, and weights for Gaussians
    :rtype: list, list, list
    """
    def weighted_mean_and_cov(x, weights):
        sum_weights = np.sum(weights)
        mean1 = []
        for i in range(x.shape[1]):
            mean1.append((np.sum(x[:, i] * weights)/sum_weights))
        mean1 = np.array(mean1)

        cov1 = np.zeros((x.shape[1], x.shape[1]))
        for i in range(x.shape[0]):
            val = x[i, :] - mean1
            cov1 += weights[i] * np.outer(val, val)
        cov1 = cov1 / sum_weights
        return mean1, cov1

    predict_kdes, obs_kdes, num_clusters = generate_output_kdes(discretization, bw_method)

    rs, r, lam_ptr = invert(discretization, bw_method)

    obs_set = discretization.get_output_observed_set()

    # Compute multivariate normal for each cluster
    means = []
    covariances = []
    cluster_weights = []
    num_obs = obs_set.check_num()

    input_dim = discretization.get_input_sample_set().get_dim()
    params = discretization.get_input_sample_set().get_values()

    for i in range(num_clusters):
        cluster_weights.append(len(np.where(obs_set.get_region() == i)[0]) / num_obs)
        if r[i] is not None:
            mean, cov = weighted_mean_and_cov(params[lam_ptr[i], :], r[i])
            means.append(mean)
            covariances.append(cov)
        else:
            means.append(None)
            covariances.append(None)

    discretization.get_input_sample_set().set_prob_type("gmm")
    discretization.get_input_sample_set().set_prob_parameters((means, covariances, cluster_weights))
    print('Diagnostic for clusters [sample average of ratios in each cluster]: ', rs)
    return means, covariances, cluster_weights


def invert_to_multivariate_gaussian(discretization, bw_method=None):
    """
    Solve the data consistent stochastic inverse problem, solving for a multivariate Gaussian.

    :param discretization: Discretization on which to perform inversion.
    :type discretization: :class:`bet.sample.discretization`
    :param bw_method: bandwidth method for `scipy.stats.gaussian_kde`.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html.
    :type bw_method: str

    :return: marginal probabilities and cluster weights
    :rtype: list, `np.ndarray`
    """
    def weighted_mean_and_cov(x, weights):
        sum_weights = np.sum(weights)
        mean1 = []
        for i in range(x.shape[1]):
            mean1.append((np.sum(x[:, i] * weights)/sum_weights))
        mean1 = np.array(mean1)

        cov1 = np.zeros((x.shape[1], x.shape[1]))
        for i in range(x.shape[0]):
            val = x[i, :] - mean1
            cov1 += weights[i] * np.outer(val, val)
        cov1 = cov1 / sum_weights
        return mean1, cov1

    predict_kdes, obs_kdes, num_clusters = generate_output_kdes(discretization, bw_method)

    rs, r, lam_ptr = invert(discretization, bw_method)

    obs_set = discretization.get_output_observed_set()

    # Compute multivariate normal
    cluster_weights = []
    num_obs = obs_set.check_num()

    params = discretization.get_input_sample_set().get_values()
    total_weights = np.zeros((discretization.get_input_sample_set().check_num(), ))

    for i in range(num_clusters):
        cluster_weights.append(len(np.where(obs_set.get_region() == i)[0]) / num_obs)
        total_weights[lam_ptr[i]] = r[i] * cluster_weights[i]
    mean, cov = weighted_mean_and_cov(params, total_weights)
    means = [mean]
    covariances = [cov]
    cluster_weights = [1.0]

    discretization.get_input_sample_set().set_prob_type("gmm")
    discretization.get_input_sample_set().set_prob_parameters((means, covariances, cluster_weights))
    print('Diagnostic for clusters [sample average of ratios in each cluster]: ', rs)
    return means, covariances, cluster_weights


def invert_to_random_variable(discretization, rv, num_reweighted=10000, bw_method=None):
    """
    Solve the data consistent stochastic inverse problem, fitting a random variable.

    `rv` can take multiple types of formats depending on type of distribution.

    A string is used for the same distribution with default parameters in each dimension.
    ex. rv = 'uniform' or rv = 'beta'

    A list or tuple of length 2 is used for the same distribution with  fixed user-defined parameters in each dimension
    as a dictionary.
    ex. rv = ['uniform', {'floc':-2, 'fscale':5}] or rv = ['beta', {'fa': 2, 'fb':5, 'floc':-2, 'fscale':5}]

    A list of length dim which entries of lists or tuples of length 2 is used for different distributions with fixed
    user-defined parameters in each dimension as a dictionary.
    ex. rv = [['uniform', {'floc':-2, 'fscale':5}],
              ['beta', {'fa': 2, 'fb':5, 'floc':-2, 'fscale':5}]]

    :param discretization: Discretization on which to perform inversion.
    :type discretization: :class:`bet.sample.discretization`
    :param rv: Type and parameters for continuous random variables.
    :type rv: str, list, or tuple
    :param num_reweighted: number of reweighted samples for fitting
    :type num_reweighted: int
    :param bw_method: bandwidth method for `scipy.stats.gaussian_kde`.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html.
    :type bw_method: str

    :return: marginal probabilities and cluster weights
    :rtype: list, `np.ndarray`
    """
    import scipy.stats as stats

    dim = discretization.get_input_sample_set().get_dim()
    if type(rv) is str:
        rv = [[rv, {}]] * dim
    elif type(rv) in (list, tuple):
        if len(rv) == 2 and type(rv[0]) is str and type(rv[1]) is dict:
            rv = [rv] * dim
        elif len(rv) != dim:
            raise bet.sample.dim_not_matching("rv has fewer entries than the dimension.")
    else:
        raise bet.sample.wrong_input("rv must be a string, list, or tuple.")

    predict_kdes, obs_kdes, num_clusters = generate_output_kdes(discretization, bw_method)

    rs, r, lam_ptr = invert(discretization, bw_method)

    obs_set = discretization.get_output_observed_set()

    # Compute multivariate normal
    cluster_weights = []
    num_obs = obs_set.check_num()

    params = discretization.get_input_sample_set().get_values()
    total_weights = np.zeros((discretization.get_input_sample_set().check_num(), ))

    for i in range(num_clusters):
        cluster_weights.append(len(np.where(obs_set.get_region() == i)[0]) / num_obs)
        total_weights[lam_ptr[i]] = r[i] * cluster_weights[i]
    total_weights = np.round(num_reweighted * total_weights/np.sum(total_weights)).astype(int)
    reweighted_vals = np.repeat(params, total_weights, axis=0)

    prob_params = []
    for i in range(dim):
        pp = [rv[i][0], {}]
        rv_continuous = getattr(stats, rv[i][0])
        A = rv_continuous.fit(reweighted_vals[:, i], **rv[i][1])
        if len(A) == 2:
            pp[1]['loc'] = A[0]
            pp[1]['scale'] = A[1]
        elif len(A) == 3:
            pp[1]['a'] = A[0]
            pp[1]['loc'] = A[1]
            pp[1]['scale'] = A[2]
        elif len(A) == 4:
            pp[1]['a'] = A[0]
            pp[1]['b'] = A[1]
            pp[1]['loc'] = A[2]
            pp[1]['scale'] = A[3]
        else:
            raise bet.sample.wrong_input("Type of random variable is not currently supported.")
        prob_params.append(pp)
    discretization.get_input_sample_set().set_prob_type('rv')
    discretization.get_input_sample_set().set_prob_parameters(prob_params)
    print('Random variable fits: ', prob_params)
    print('Diagnostic for clusters [sample average of ratios in each cluster]: ', rs)
    return prob_params



# Copyright (C) 2014-2020 The BET Development Team

"""
This module provides methods for Voronoi plots.
"""

import copy
import math
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from bet.Comm import comm, MPI
import bet.sample as sample


class dim_not_matching(Exception):
    """
    Exception for when the dimension is inconsistent.
    """


class bad_object(Exception):
    """
    Exception for when the wrong type of object is used.
    """


class missing_attribute(Exception):
    """
    Exception for missing attribute.
    """


def plot_1D_voronoi(sample_set, density=True, filename="file",
                    lam_ref=None, interactive=False,
                    lambda_label=None, file_extension=".png"):
    """
    This makes a 1d Voronoi plot of the input probability measure for a
    1D Voronoi sample set. If the sample_set object is a discretization
    object, we assume that the probabilities to be plotted are from
    the input space.

    .. note::

        Do not specify the file extension in the file name.

    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :param density: Plot prob. density instead of prob. measure.
    :type density: bool
    :param filename: Prefix for output files.
    :type filename: str
    :param lam_ref: Reference parameters.
    :type lam_ref: :class:`~numpy.ndarray` of shape (ndim,) or None
    :param interactive: Whether or not to display interactive plots.
    :type interactive: bool
    :param lambda_label: Label for each parameter for plots.
    :type lambda_label: list of length nbins of strings or None
    :param string file_extension: file extenstion

    """

    # Check inputs
    if isinstance(sample_set, sample.discretization):
        sample_obj = sample_set._input_sample_set
        if sample_obj is None:
            raise missing_attribute("Missing input_sample_set")
    elif isinstance(sample_set, sample.sample_set_base):
        sample_obj = sample_set
    else:
        raise bad_object("Improper sample object")

    if sample_obj._dim != 1:
        raise dim_not_matching("Only applicable for 1D domains.")

    # Check for global probabilities
    if sample_obj._probabilities is None:
        if sample_obj._probabilities_local is None:
            raise missing_attribute("Missing probabilities")
        else:
            sample_obj.local_to_global()

    if lam_ref is None:
        lam_ref = sample_obj._reference_value

    # Form 1D Voronoi
    ind_sort = np.argsort(sample_obj._values, axis=0)
    ends = 0.5 * (sample_obj._values[ind_sort]
                  [1::] + sample_obj._values[ind_sort][0:-1])
    ends = ends[:, 0, 0]
    mins = np.array([sample_obj._domain[0][0]] + list(ends))
    maxes = np.array(list(ends) + [sample_obj._domain[0][1]])

    # Make plot
    if comm.rank == 0:
        fig = plt.figure(0, constrained_layout=True)
        if density:
            plt.hlines(
                sample_obj._probabilities[ind_sort] / (maxes - mins), mins, maxes)
            plt.ylabel(r'$\rho_{\lambda}$', fontsize=20)
        else:
            plt.hlines(sample_obj._probabilities[ind_sort], mins, maxes)
            plt.ylabel(r'$P_{\Lambda}(\mathcal{V}_i)$', fontsize=20)

        if lam_ref is not None:
            plt.plot(lam_ref[0], 0.0, 'ko', markersize=10)

        if lambda_label is None:
            label1 = r'$\lambda$'
        else:
            label1 = lambda_label[0]
        plt.xlabel(label1, fontsize=20)

        fig.savefig(filename + file_extension)
        if interactive:
            plt.show()


def plot_2D_voronoi(sample_set, density=True, colormap_type='BuGn',
                    filename="file",
                    lam_ref=None, interactive=False,
                    lambda_label=None, file_extension=".png"):
    """
    This makes a 2D Voronoi plot of the input probability measure for a
    2D Voronoi sample set. If the sample_set object is a discretization
    object, we assume that the probabilities to be plotted are from
    the input space.

    .. note::

        Do not specify the file extension in the file name.

    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :param density: Plot prob. density instead of prob. measure.
    :type density: bool
    :param colormap_type: type of color map to use
    :type colormap_type: str
    :param filename: Prefix for output files.
    :type filename: str
    :param lam_ref: Reference parameters.
    :type lam_ref: :class:`~numpy.ndarray` of shape (ndim,) or None
    :param interactive: Whether or not to display interactive plots.
    :type interactive: bool
    :param lambda_label: Label for each parameter for plots.
    :type lambda_label: list of length nbins of strings or None
    :param string file_extension: file extenstion

    """

    from scipy.spatial import Voronoi, voronoi_plot_2d

    # Check inputs
    if isinstance(sample_set, sample.discretization):
        sample_obj = sample_set._input_sample_set
        if sample_obj is None:
            raise missing_attribute("Missing input_sample_set")
    elif isinstance(sample_set, sample.sample_set_base):
        sample_obj = sample_set
    else:
        raise bad_object("Improper sample object")

    if sample_obj._dim != 2:
        raise dim_not_matching("Only applicable for 2D domains.")

    # Check for global probabilities
    if sample_obj._probabilities is None:
        if sample_obj._probabilities_local is None:
            raise missing_attribute("Missing probabilities")
        else:
            sample_obj.local_to_global()
    if sample_obj._values is None:
        sample_obj.local_to_global()

    if lam_ref is None:
        lam_ref = sample_obj._reference_value

    # Form Voronoi
    if comm.rank == 0:
        vor = Voronoi(sample_obj._values)
        regions, vertices = voronoi_finite_polygons_2d(vor)
        points = sample_obj._values

        # Make plot
        fig = plt.figure(0, constrained_layout=True)
        cmap = matplotlib.cm.get_cmap(colormap_type)
        if density:
            P = sample_obj._probabilities / sample_obj._volumes
        else:
            P = sample_obj._probabilities
        P_max = np.max(P)

        # plot each cell
        for i, region in enumerate(regions):
            polygon = vertices[region]
            plt.fill(*list(zip(*polygon)),
                     color=cmap(P[i] / P_max), edgecolor='k', linewidth=0.005)

        plt.axis([sample_obj._domain[0][0], sample_obj._domain[0][1],
                  sample_obj._domain[1][0], sample_obj._domain[1][1]])
        if lam_ref is not None:
            plt.plot(lam_ref[0], lam_ref[1], 'ro', markersize=10)

        if lambda_label is None:
            label1 = r'$\lambda_1$'
            label2 = r'$\lambda_2$'
        else:
            label1 = lambda_label[0]
            label1 = lambda_label[1]
        plt.xlabel(label1, fontsize=20)
        plt.ylabel(label2, fontsize=20)

        # plot colorbar
        ax, _ = matplotlib.colorbar.make_axes(plt.gca(), shrink=0.9)
        if density:
            cbar = matplotlib.colorbar.ColorbarBase(ax, cmap=cmap,
                                                    norm=matplotlib.colors.Normalize(vmin=0.0, vmax=P_max), label=r'$\rho_{\Lambda}$')
        else:
            cbar = matplotlib.colorbar.ColorbarBase(ax, cmap=cmap,
                                                    norm=matplotlib.colors.Normalize(vmin=0.0, vmax=P_max), label=r'$P_{\Lambda}(\mathcal{V}_i)$')
        text = cbar.ax.yaxis.label
        font = matplotlib.font_manager.FontProperties(size=20)
        text.set_font_properties(font)

        fig.savefig(filename + file_extension)
        if interactive:
            plt.show()


def voronoi_finite_polygons_2d(vor, radius=None):
    """
    Reconstruct infinite Voronoi regions in a 2D diagram to finite
    regions.

    :param vor: Voronoi input diagram
    :type vor: :class:`scipy.spatial.Voronoi`
    :param radius: Distance to points at infinity. Optional.
    :type radius: float
    :param regions: Indices of vertices in each revised Voronoi regions.
    :type regions: list of tuples
    :param vertices: Coordinates for revised Voronoi vertices.
    Same as coordinates
    of input vertices, with 'points at infinity' appended to the end.
    :type vertices: list of tuples
    :rtype: tuple
    :returns (regions, vertices)

    """

    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge

            t = vor.points[p2] - vor.points[p1]  # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains general tools for BET including saving and loading objects, and reshaping objects. The most
important methods are:

* :mod:`~bet.util.get_global_values` concatenates local arrays into global arrays.
* :mod:`~bet.util.save_object` saves all types of objects.
* :mod:`~bet.util.load_object` loads all types of saved objects.
* :mod:`~bet.util.load_object_parallel` loads all types of saved parallel objects.

"""

import sys
import collections.abc
import os
import glob
import logging
import numpy as np
import bet.sample
from bet.Comm import comm, MPI

possible_types = {int: MPI.INT, float: MPI.DOUBLE}


def meshgrid_ndim(X):
    """
    Return coordinate matrix from two or more coordinate vectors.
    Handles a maximum of 10 vectors.

    Make N-D coordinate arrays for vectorized evaluations of
    N-D scalar/vector fields over N-D grids, given
    one-dimensional coordinate arrays (x1, x2,..., xn).


    :param X: A tuple containing the 1d coordinate arrays
    :type X: tuple
    :rtype: :class:`~numpy.ndarray` of shape (num_grid_points,n)
    :returns: X_new
    """
    n = len(X)
    alist = []
    for i in range(n):
        alist.append(X[i])
    for i in range(n, 10):
        alist.append(np.array([0]))

    a, b, c, d, e, f, g, h, i, j = np.meshgrid(alist[0],
                                               alist[1],
                                               alist[2],
                                               alist[3],
                                               alist[4],
                                               alist[5],
                                               alist[6],
                                               alist[7],
                                               alist[8],
                                               alist[9],
                                               indexing='ij')

    X_new = np.vstack(
        (a.flat[:],
         b.flat[:],
         c.flat[:],
         d.flat[:],
         e.flat[:],
         f.flat[:],
         g.flat[:],
         h.flat[:],
         i.flat[:],
         j.flat[:])).transpose()
    X_new = X_new[:, 0:n]

    return X_new


def get_global_values(array, shape=None):
    """
    Concatenates local arrays into global array using :meth:`np.vstack`.

    :param array: Array.
    :type P_samples: :class:`~numpy.ndarray`
    :rtype: :class:`~numpy.ndarray`
    :returns: array
    """
    if comm.size == 1:
        return array
    else:
        # Figure out the subtype of the elements of the array
        dtype = array.dtype
        mpi_dtype = False
        for ptype in possible_types.keys():
            # suppress FutureWarning
            if ptype is int or ptype is MPI.INT:
                comp_type = np.integer
            elif ptype is float or ptype is MPI.FLOAT:
                comp_type = np.floating

            if np.issubdtype(dtype, comp_type):
                mpi_dtype = True
                dtype = ptype

        if shape is None or not mpi_dtype:
            # do a lowercase allgather
            a_shape = len(array.shape)
            array = comm.allgather(array)
            if a_shape <= 1:
                return np.hstack(array)
            else:
                return np.vstack(array)
        else:
            # do an uppercase Allgather
            whole_a = np.empty(shape, dtype=dtype)
            comm.Allgather([array.ravel(), possible_types[dtype]], [whole_a,
                                                                    possible_types[dtype]])
            return whole_a


def fix_dimensions_vector(vector):
    """
    Fix the dimensions of an input so that it is a :class:`numpy.ndarray` of
    shape (N,).
    :param vector: numerical object
    :rtype: :class:`numpy.ndarray`
    :returns: array of shape (N,)
    """
    if not isinstance(vector, collections.abc.Iterable):
        vector = np.array([vector])
    elif not isinstance(vector, np.ndarray):
        vector = np.array(vector)
    return vector.flat[:]


def fix_dimensions_vector_2darray(vector):
    """
    Fix the dimensions of an input so that it is a :class:`numpy.ndarray` of
    shape (N,1).

    :param vector: numerical object
    :rtype: :class:`numpy.ndarray`
    :returns: array of shape (N,1)

    """
    if not isinstance(vector, collections.abc.Iterable):
        vector = np.array([vector])
    elif not isinstance(vector, np.ndarray):
        vector = np.array(vector)
    if len(vector.shape) <= 1:
        vector = np.expand_dims(vector, axis=1)
    return vector


def fix_dimensions_domain(domain):
    """
    Fix the dimensions of an input so that it is a :class:`numpy.ndarray` of
    shape (dim, 2).

    :param vector: numerical object of at least length 2
    :type vector: :class:`collections.abc.Iterable`
    :rtype: :class:`numpy.ndarray`
    :retuns: array of shape (dim, 2)

    """
    if not isinstance(domain, np.ndarray):
        if len(domain) == 2:
            domain = np.expand_dims(domain, axis=0)
        else:
            raise TypeError("The length must be at least 2.")
    elif len(domain.shape) == 1 and domain.shape[0] == 2:
        domain = np.expand_dims(domain, axis=0)
    elif len(domain.shape) == 2 and domain.shape[1] == 2:
        pass  # The shape is already correct!
    elif len(domain.shape) == 2 and domain.shape[0] == 2:
        domain = domain.transpose()
    else:
        raise TypeError("At least one dimension must have a length of 2.")
    return domain


def fix_dimensions_data(data, dim=None):
    """
    Fix the dimensions of an input so that it is a :class:`numpy.ndarray` of
    shape (N, dim).

    If ``dim`` is non-specified:
    If ``data`` is a non-iterable number assumes that ``dim==1``.
    If ``data`` is a numpy array with len(shape) == 1 assumes that ``dim==1``.
    If ``data`` is a numpy array with len(shape) == 2 assumes that
    ``dim==shape[1]``.


    :param data: numerical object
    :param int dim: The dimension of the "data" space.
    :rtype: :class:`numpy.ndarray`
    :returns: array of shape (N, dim)

    """
    if dim is None:
        if not isinstance(data, np.ndarray):
            return fix_dimensions_vector_2darray(data)
        elif len(data.shape) == 1:
            return fix_dimensions_vector_2darray(data)
        else:
            return data

    data = fix_dimensions_vector_2darray(data)
    if len(data.shape) > 1 and data.shape[1] != dim:
        return data.transpose()
    else:
        return data


def clean_data(data):
    """
    Clean data so that NaN->0, inf-> maxfloat, -inf-> -maxfloat

    :param data: numerical object
    :type data: :class:`numpy.ndarray`
    :rtype: :class:`numpy.ndarray`
    :returns: array of shape (data.shape)

    """
    data[np.isnan(data)] = 0.0
    data[np.isinf(data)] = np.sign(data[np.isinf(data)]) * sys.float_info[0]

    return data


def save_object(save_set, file_name, globalize=True):
    """
    Save BET object.

    :param save_set: Object to Save.
    :param file_name: Filename to save to.
    :type file_name: str
    :param globalize: Whether or not to globalize parallel objects.
    :type globalize: bool
    """
    import pickle
    # create processor specific file name
    if comm.size > 1 and not globalize:
        local_file_name = os.path.join(os.path.dirname(file_name),
                                       "proc{}_{}".format(comm.rank,
                                                          os.path.basename(file_name)))
    else:
        local_file_name = file_name
    if os.path.exists(local_file_name + '.p'):
        logging.warn("Warning! Output file already exists. New object will be appended.")
    # globalize
    if globalize:
        save_set.local_to_global()
    comm.barrier()
    pickle.dump(save_set, open(local_file_name + '.p', "wb"))
    comm.barrier()
    return local_file_name


def load_object(file_name, localize=False):
    """
    Load saved objects.

    :param file_name: Filename of object.
    :type file_name: str
    :param localize: Whether or not to localize parallel object.
    :type localize: bool
    :return: The saved object
    """
    import pickle
    # check to see if parallel file name
    if file_name.startswith('proc_'):
        # logging.warning("Avoid starting filenames with 'proc_'. Unable to localize.")
        localize = False
    elif not os.path.exists(file_name+'.p') and os.path.exists('proc0_'+file_name+'.p'):
        return load_object_parallel(file_name)
    loaded_set = pickle.load(open(file_name+'.p', "rb"))
    if localize:
        loaded_set.global_to_local()
    return loaded_set


def load_object_parallel(file_name):
    """
    Load saved paralell objects.

    :param file_name: Filename of object.
    :type file_name: str
    :return: The saved object

    """
    save_dir = os.path.dirname(file_name)
    base_name = os.path.basename(file_name)
    files = glob.glob(os.path.join(save_dir, "proc*_{}".format(base_name+'.p')))
    if len(files) == comm.size:
        logging.info("Loading sample set using parallel files (same nproc)")
        # if the number of processors is the same then set mdat to
        # be the one with the matching processor number (doesn't
        # really matter)
        local_file_name = os.path.join(os.path.dirname(file_name),
                                       "proc{}_{}".format(comm.rank,
                                                          os.path.basename(file_name)))
        return load_object(local_file_name)
    else:
        raise bet.sample.dim_not_matching("Number of parallel files is different from nproc.")
    # SM possibly re-add the feature to have different numbers. Probably not necessary.



# Copyright (C) 2014-2020 The BET Development Team

r"""
This subpackage contains

* :class:`~bet.postProcess.plotP` plots :math:`P` and/or volumes (:math:`\mu`) of Voronoi cells
* :class:`~bet.postProcess.plotDomains` plots the data domain :math:`\mathcal{D}` in 2D
* :class:`~bet.postProcess.postTools` has tools for postprocessing
* :class:`~bet.postProcess.compareP` has tools for comparing measures
"""
__all__ = ['plotP', 'plotDomains', 'postTools', 'plotVoronoi', 'compareP']



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains unittests for :mod:`~bet.calculateP.calculateR`
"""

import unittest
import os
import pyDOE
import numpy.testing as nptest
import numpy as np
import scipy.io as sio
import bet
import bet.sampling.basicSampling as bsam
from bet.Comm import comm
import bet.sample
from bet.sample import sample_set
from bet.sample import discretization as disc
import collections
from test.problem_setups import *


class Test_calculateR(unittest.TestCase):
    """
    Testing ``bet.calculateP.calculateR``
    """
    def setUp(self):
        self.in_dim = 1
        self.out_dim = 1
        self.vals = np.ones((10, ))
        self.vals_marg= np.ones((10, ))

    def test_kde(self):
        """
        Test ``bet.calculateP.calculateR.invert_to_kde``
        """
        disc, _ = random_kde(dim=self.in_dim, out_dim=self.out_dim, level=2)
        disc.get_input_sample_set().pdf(self.vals)
        disc.get_input_sample_set().pdf_init(self.vals)
        disc.get_input_sample_set().marginal_pdf(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf(self.vals_marg, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals_marg, i=0)

    def test_rv(self):
        """
        Test ``bet.calculateP.calculateR.invert_to_random_variable``
        """
        disc, _ = random_rv(dim=self.in_dim, out_dim=self.out_dim, level=2)
        disc.get_input_sample_set().pdf(self.vals)
        disc.get_input_sample_set().pdf_init(self.vals)
        disc.get_input_sample_set().marginal_pdf(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf(self.vals_marg, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals_marg, i=0)

    def test_gmm(self):
        """
        Test ``bet.calculateP.calculateR.invert_to_gmm``
        """
        disc, _ = random_gmm(dim=self.in_dim, out_dim=self.out_dim, level=2)
        disc.get_input_sample_set().pdf(self.vals)
        disc.get_input_sample_set().pdf_init(self.vals)
        disc.get_input_sample_set().marginal_pdf(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf(self.vals_marg, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals_marg, i=0)

    def test_multivariate_gaussian(self):
        """
        Test ``bet.calculateP.calculateR.invert_to_multivariate_gaussian``
        """
        disc, _ = random_multivariate_gaussian(dim=self.in_dim, out_dim=self.out_dim, level=2)
        disc.get_input_sample_set().pdf(self.vals)
        disc.get_input_sample_set().pdf_init(self.vals)
        disc.get_input_sample_set().marginal_pdf(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals, i=0)
        disc.get_input_sample_set().marginal_pdf(self.vals_marg, i=0)
        disc.get_input_sample_set().marginal_pdf_init(self.vals_marg, i=0)

class Test_calculateR_3to2(Test_calculateR):
    """
    Testing ``bet.calculateP.calculateR`` with a 3 to 2 map.
    """
    def setUp(self):
        self.in_dim = 3
        self.out_dim = 3
        self.vals = np.ones((10, 3))
        self.vals_marg= np.ones((10, ))


class Test_invert_to_random_variable(unittest.TestCase):
    """
    Test `bet.calculateP.calculateR.invert_to_random_variable`
    """
    def test_string(self):
        """
        Test when rv is a string.
        """
        random_rv(dim=2, out_dim=1, rv_invert='beta', level=2)

    def test_list1(self):
        """
        Test when rv is a list of length 2.
        """
        random_rv(dim=2, out_dim=1, rv_invert=['beta', {'loc': 0.25}], level=2)

    def test_list2(self):
        """
        Test when rv is a list of lists.
        """
        random_rv(dim=2, out_dim=1, rv_invert=[['beta', {'floc': 0.25}], ['norm', {}]], level=2)

    def test_sample_from_updated(self):
        disc, _ = random_rv(dim=2, out_dim=1, rv_invert=[['beta', {'floc': 0.25}], ['norm', {}]], level=2)
        new_set = bsam.sample_from_updated(disc, num_samples=100)
        assert new_set.get_dim() == 2
        assert new_set.check_num() == 100

        disc, _ = random_gmm(dim=2, out_dim=1, level=2)
        new_set = bsam.sample_from_updated(disc, num_samples=100)
        assert new_set.get_dim() == 2
        assert new_set.check_num() == 100

        disc, _ = random_kde(dim=2, out_dim=1, level=2)
        new_set = bsam.sample_from_updated(disc, num_samples=100)
        assert new_set.get_dim() == 2
        assert new_set.check_num() == 100
        disc.global_to_local()




class Test_rejection_sampling(unittest.TestCase):
    def Test_rejection_sampling(self):
        """
        Testing ``bet.calculateP.calculateR.invert_rejection_sampling``
        """
        rv = 'uniform'
        dim = 1
        out_dim = 1
        num_samples = 1000
        globalize = True
        rv2 = "norm"
        def my_model(samples):
            A = np.eye(dim, out_dim)
            return np.dot(samples, A)

        sampler1 = bsam.sampler(lb_model=my_model, error_estimates=False, jacobians=False)
        sampler1.random_sample_set(rv, dim, num_samples, globalize)
        disc1 = sampler1.compute_qoi_and_create_discretization()

        sampler2 = bsam.sampler(lb_model=my_model, error_estimates=False, jacobians=False)
        sampler2.random_sample_set(rv2, dim, num_samples, globalize)
        disc2 = sampler1.compute_qoi_and_create_discretization()

        disc1.set_output_observed_set(disc2.get_output_sample_set())
        calculateR.invert_rejection_sampling(disc1)









# Copyright (C) 2014-2020 The BET Development Team

# -*- coding: utf-8 -*-
import numpy as np
import math as m

'''
Suggested changes for user:

Try setting QoI_num = 2.

Play around with the x1, y1, and/or, x2, y2 values to try and
"optimize" the QoI to give the highest probability region
on the reference parameter above.

Hint: Try using QoI_num = 1 and systematically varying the
x1 and y1 values to find QoI with contour structures (as inferred
through the 2D marginal plots) that are nearly orthogonal.

Some interesting pairs of QoI to compare are:
(x1,y1)=(0.5,0.5) and (x2,y2)=(0.25,0.25)
(x1,y1)=(0.5,0.5) and (x2,y2)=(0.15,0.15)
(x1,y1)=(0.5,0.5) and (x2,y2)=(0.25,0.15)
'''
# Choose the number of QoI
QoI_num = 2

# Specify the spatial points to take measurements of solution defining the QoI
if QoI_num == 1:
    x1 = 0.5
    y1 = 0.5
    x = np.array([x1])
    y = np.array([y1])
else:
    x1 = 0.1
    y1 = 0.05
    x2 = 0.05
    y2 = 0.1
    x = np.array([x1, x2])
    y = np.array([y1, y2])


class QoI_component(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def eval(self, parameter_samples):
        if parameter_samples.shape == (2,):
            lam1 = parameter_samples[0]
            lam2 = parameter_samples[1]
        else:
            lam1 = parameter_samples[:, 0]
            lam2 = parameter_samples[:, 1]
        z = np.sin(m.pi * self.x * lam1) * np.sin(m.pi * self.y * lam2)
        return z


# Specify the QoI maps
if QoI_num == 1:
    def QoI_map(parameter_samples):
        Q1 = QoI_component(x[0], y[0])
        return np.array([Q1.eval(parameter_samples)]).transpose()
else:
    def QoI_map(parameter_samples):
        Q1 = QoI_component(x[0], y[0])
        Q2 = QoI_component(x[1], y[1])
        return np.array([Q1.eval(parameter_samples),
                         Q2.eval(parameter_samples)]).transpose()

# Define a model that is the QoI map


def my_model(parameter_samples):
    QoI_samples = QoI_map(parameter_samples)
    return QoI_samples



# Copyright (C) 2014-2020 The BET Development Team

import bet.calculateP.calculateP as calcP
import bet.calculateP.simpleFunP as sfun
import bet.sampling.basicSampling as bsam
import numpy as np
import scipy.io as sio
import bet.sample as sample

# Import "Truth"
mdat = sio.loadmat('Q_3D')
Q = mdat['Q']
Q_ref = mdat['Q_true']

# Import Data
samples = mdat['points'].transpose()
lam_domain = np.array([[-900, 1200], [0.07, .15], [0.1, 0.2]])

# Create input, output, and discretization from data read from file
points = mdat['points']
input_sample_set = sample.sample_set(points.shape[0])
input_sample_set.set_values(points.transpose())
input_sample_set.set_domain(lam_domain)
print("Finished loading data")


def postprocess(station_nums, ref_num):

    filename = 'P_q' + str(station_nums[0] + 1) + \
        '_q' + str(station_nums[1] + 1)
    if len(station_nums) == 3:
        filename += '_q' + str(station_nums[2] + 1)
    filename += '_ref_' + str(ref_num + 1)

    data = Q[:, station_nums]
    output_sample_set = sample.sample_set(data.shape[1])
    output_sample_set.set_values(data)
    q_ref = Q_ref[ref_num, station_nums]

    # Create Simple function approximation
    # Save points used to parition D for simple function approximation and the
    # approximation itself (this can be used to make close comparisions...)

    output_probability_set = sfun.regular_partition_uniform_distribution_rectangle_scaled(
        output_sample_set, q_ref, rect_scale=0.15,
        cells_per_dimension=np.ones((data.shape[1],)))
    my_disc = sample.discretization(input_sample_set, output_sample_set,
                                    output_probability_set)

    # Calclate P on the actual samples with assumption that voronoi cells have
    # equal size
    input_sample_set.estimate_volume_mc()
    print("Calculating prob")
    calcP.prob(my_disc)

# Post-process and save P and emulated points
ref_num = 14

# q1, q5, q2 ref 15
station_nums = [0, 4, 1]  # 1, 5, 2
postprocess(station_nums, ref_num)

"""
# q1, q5 ref 15
station_nums = [0, 4] # 1, 5
postprocess(station_nums, ref_num)

# q1, q5, q12 ref 16
#ref_num = 15
station_nums = [0, 4, 11] # 1, 5, 12
postprocess(station_nums, ref_num)


station_nums = [0, 8, 6] # 1, 5, 12
postprocess(station_nums, ref_num)


station_nums = [0, 8, 11] # 1, 5, 12
postprocess(station_nums, ref_num)
"""



#!/usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

'''
The python script for building the BET package and subpackages.
'''
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(name='bet',
      version='3.0.0',
      description='A toolkit for data-consistent stochastic problems.',
      author='Steven Mattis',
      author_email='steve.a.mattis@gmail.com',
      license='GNU LGPL',
      url='https://github.com/UT-CHG/BET',
      packages=['bet',
                'bet.sampling',
                'bet.calculateP',
                'bet.postProcess',
                'bet.sensitivity'],
      install_requires=['matplotlib',
                        'pyDOE',
                        'numpy',
                        'scipy',
                        'pytest',
                        'mpi4py'])



# Copyright (C) 2014-2020 The BET Development Team

"""
This package contains all of the tests for :program:`BET`. The package
structure mirrors the ``bet`` package structure.
"""
__all__ = ['test_calculateP', 'test_simpleFunP',
           'test_calculateError', 'test_calculateR']



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains tests for :module:`bet.sensitivity.gradients`.
Most of these tests should make sure certain values are within a tolerance
rather than exact due to machine precision.
"""
import unittest
import bet.sensitivity.gradients as grad
import numpy as np
import numpy.testing as nptest
import bet.sample as sample


class GradientsMethods:
    """
    Test all methods in :module:`bet.sensitivity.gradients`
    """
    # Test sampling methods

    def test_sample_lp_ball(self):
        """
        Test :meth:`bet.sensitivity.gradients.sample_lp_ball`.
        """
        self.cluster_set = grad.sample_lp_ball(self.input_set_centers,
                                               self.num_close, self.rvec)

        # Test the method returns the correct dimensions
        self.assertEqual(self.cluster_set._values.shape, ((self.num_close + 1) *
                                                          self.num_centers, self.input_dim))

        # Check the method returns centers followed by the clusters around the
        # first center.
        self.repeat = np.repeat(self.centers, self.num_close, axis=0)

        nptest.assert_array_less(np.linalg.norm(self.cluster_set._values[
            self.num_centers:] - self.repeat, 2, axis=1),
            np.max(self.rvec))

        # Check that the samples are in lam_domain
        self.cluster_set.update_bounds()
        left = np.all(np.greater_equal(self.cluster_set._values,
                                       self.cluster_set._left))
        right = np.all(np.less_equal(self.cluster_set._values,
                                     self.cluster_set._right))
        assert np.all(np.logical_and(left, right))

    def test_sample_linf_ball(self):
        """
        Test :meth:`bet.sensitivity.gradients.sample_linf_ball`.
        """
        self.cluster_set = grad.sample_linf_ball(self.input_set_centers,
                                                 self.num_close, self.rvec)

        # Test the method returns the correct dimensions
        self.assertEqual(self.cluster_set._values.shape, ((self.num_close + 1) *
                                                          self.num_centers, self.input_dim))

        # Check the method returns centers followed by the clusters around the
        # first center.
        self.repeat = np.repeat(self.centers, self.num_close, axis=0)

        nptest.assert_array_less(np.linalg.norm(self.cluster_set._values[
            self.num_centers:] - self.repeat, np.inf, axis=1),
            np.max(self.rvec))

        # Check that the samples are in lam_domain
        self.cluster_set.update_bounds()
        left = np.all(np.greater_equal(self.cluster_set._values,
                                       self.cluster_set._left))
        right = np.all(np.less_equal(self.cluster_set._values,
                                     self.cluster_set._right))
        assert np.all(np.logical_and(left, right))

    def test_sample_l1_ball(self):
        """
        Test :meth:`bet.sensitivity.gradients.sample_l1_ball`.
        """
        self.cluster_set = grad.sample_l1_ball(self.input_set_centers,
                                               self.num_close, self.rvec)

        # Test that the samples are within max(rvec) of center (l1 dist)
        self.repeat = np.repeat(self.centers, self.num_close, axis=0)
        nptest.assert_array_less(np.linalg.norm(self.cluster_set._values[
            self.num_centers:] - self.repeat, 1, axis=1),
            np.max(self.rvec))

        # Test the method returns the correct dimensions
        self.assertEqual(self.cluster_set._values.shape, ((self.num_close + 1) *
                                                          self.num_centers, self.input_dim))

    # Test FD methods
    def test_pick_ffd_points(self):
        """
        Test :meth:`bet.sensitivity.gradients.sample_linf_ball`.
        """
        self.cluster_set = grad.pick_ffd_points(self.input_set_centers,
                                                self.rvec)

        #self.samples = grad.pick_ffd_points(self.centers, self.rvec)

        if not isinstance(self.rvec, np.ndarray):
            self.rvec = np.ones(self.input_dim) * self.rvec

        # Check the distance to the corresponding center is equal to rvec
        self.centersrepeat = np.repeat(
            self.centers, self.cluster_set._dim, axis=0)
        nptest.assert_array_almost_equal(np.linalg.norm(self.centersrepeat -
                                                        self.cluster_set._values[self.num_centers:], axis=1),
                                         np.tile(self.rvec, self.num_centers))

        # Test the method returns the correct dimensions
        self.assertEqual(self.cluster_set._values.shape, ((self.cluster_set._dim +
                                                           1) * self.num_centers, self.cluster_set._dim))

    def test_pick_cfd_points(self):
        """
        Test :meth:`bet.sensitivity.gradients.sample_l1_ball`.
        """
        self.cluster_set = grad.pick_cfd_points(self.input_set_centers,
                                                self.rvec)

        if not isinstance(self.rvec, np.ndarray):
            self.rvec = np.ones(self.input_dim) * self.rvec

        # Check the distance to the corresponding center is equal to rvec
        self.centersrepeat = np.repeat(self.centers, 2 * self.cluster_set._dim,
                                       axis=0)
        nptest.assert_array_almost_equal(np.linalg.norm(self.centersrepeat -
                                                        self.cluster_set._values[self.num_centers:], axis=1),
                                         np.tile(self.rvec, self.num_centers * 2))

        # Test the method returns the correct dimension
        self.assertEqual(self.cluster_set._values.shape, ((2 * self.input_dim + 1)
                                                          * self.num_centers, self.cluster_set._dim))

    # Test RBF methods
    def test_radial_basis_function(self):
        """
        Test :meth:`bet.sensitivity.gradients.radial_basis_function`.
        """
        self.rbf = grad.radial_basis_function(self.radii_rbf, 'Multiquadric')

        # Test the method returns the correct number of function evaluations
        self.assertEqual(self.rbf.shape, (self.radii_rbf.shape))

        # Test the method returns an error when kernel is not available
        with self.assertRaises(ValueError):
            grad.radial_basis_function(self.radii_rbf, 'DNE')

    def test_radial_basis_function_dxi(self):
        """
        Test :meth:`bet.sensitivity.gradients.radial_basis_function_dxi`.
        """
        self.rbfdxi = grad.radial_basis_function_dxi(self.radii_rbfdxi,
                                                     self.dxi, 'Multiquadric')

        # Test the method returns the correct number of function evaluations
        self.assertEqual(self.rbfdxi.shape, (self.radii_rbfdxi.shape))

        # Test the method returns an error when kernel is not available
        with self.assertRaises(ValueError):
            grad.radial_basis_function_dxi(self.radii_rbfdxi, self.dxi, 'DNE')

    # Test gradient approximation methods
    def test_calculate_gradients_rbf(self):
        """
        Test :meth:`bet.sensitivity.gradients.calculate_gradients_rbf`.
        """
        self.output_set = sample.sample_set(self.output_dim)
        self.cluster_set = grad.sample_l1_ball(self.input_set_centers,
                                               self.num_close, self.rvec)
        num_centers = self.input_set_centers.check_num()
        self.output_set.set_values(self.cluster_set._values.dot(self.coeffs))
        self.cluster_disc = sample.discretization(self.cluster_set,
                                                  self.output_set)

        self.center_disc = grad.calculate_gradients_rbf(self.cluster_disc,
                                                        num_centers)

        # Test the method returns the correct size tensor
        self.jacobians = self.center_disc._input_sample_set._jacobians
        self.assertEqual(self.jacobians.shape, (self.num_centers,
                                                self.output_dim, self.input_dim))

        # Test that each vector is normalized or a zero vector
        normG = np.linalg.norm(self.jacobians, ord=1, axis=2)

        # If its a zero vectors, make it the unit vector in input_dim
        self.jacobians[normG == 0] = 1.0 / self.input_dim
        nptest.assert_array_almost_equal(np.linalg.norm(self.jacobians, ord=1,
                                                        axis=2), np.ones((self.jacobians.shape[0],
                                                                          self.jacobians.shape[1])))

    def test_calculate_gradients_ffd(self):
        """
        Test :meth:`bet.sensitivity.gradients.calculate_gradients_ffd`.
        """
        self.output_set = sample.sample_set(self.output_dim)
        self.cluster_set = grad.pick_ffd_points(self.input_set_centers,
                                                self.rvec)
        num_centers = self.input_set_centers.check_num()
        self.output_set.set_values(self.cluster_set._values.dot(self.coeffs))
        self.cluster_disc = sample.discretization(self.cluster_set,
                                                  self.output_set)

        self.center_disc = grad.calculate_gradients_ffd(self.cluster_disc)
        self.jacobians = self.center_disc._input_sample_set._jacobians

        # Test the method returns the correct size tensor
        self.assertEqual(self.jacobians.shape,
                         (self.num_centers, self.output_dim, self.input_dim))

        # Test that each vector is normalized or a zero vector
        normG = np.linalg.norm(self.jacobians, ord=1, axis=2)

        # If its a zero vectors, make it the unit vector in input_dim
        self.jacobians[normG == 0] = 1.0 / self.input_dim
        nptest.assert_array_almost_equal(np.linalg.norm(self.jacobians, ord=1,
                                                        axis=2), np.ones((self.jacobians.shape[0],
                                                                          self.jacobians.shape[1])))

    def test_calculate_gradients_cfd(self):
        """
        Test :meth:`bet.sensitivity.gradients.calculate_gradients_cfd`.
        """
        self.output_set = sample.sample_set(self.output_dim)
        self.cluster_set = grad.pick_cfd_points(self.input_set_centers,
                                                self.rvec)
        num_centers = self.input_set_centers.check_num()
        self.output_set.set_values(self.cluster_set._values.dot(self.coeffs))
        self.cluster_disc = sample.discretization(self.cluster_set,
                                                  self.output_set)

        self.center_disc = grad.calculate_gradients_cfd(self.cluster_disc)
        self.jacobians = self.center_disc._input_sample_set._jacobians

        # Test the method returns the correct size tensor
        self.assertEqual(self.jacobians.shape,
                         (self.num_centers, self.output_dim, self.input_dim))

        # Test that each vector is normalized or a zero vector
        normG = np.linalg.norm(self.jacobians,
                               ord=1, axis=2)

        # If its a zero vectors, make it the unit vector in input_dim
        self.jacobians[normG == 0] = 1.0 / self.input_dim
        nptest.assert_array_almost_equal(np.linalg.norm(self.jacobians, ord=1,
                                                        axis=2), np.ones((self.jacobians.shape[0],
                                                                          self.jacobians.shape[1])))

# Test the accuracy of the gradient approximation methods


class GradientsAccuracy:
    """
    Test the accuracy of the gradient approximation method in
        :module:`bet.sensitivity.gradients`
    """

    def test_calculate_gradients_rbf_accuracy(self):
        """
        Test :meth:`bet.sensitivity.gradients.calculate_gradients_rbf`.
        """
        self.center_disc = grad.calculate_gradients_rbf(
            self.cluster_disc_rbf, normalize=False)
        self.jacobians = self.center_disc._input_sample_set._jacobians

        nptest.assert_allclose(self.jacobians - self.G_exact, 0,
                               atol=2)

    def test_calculate_gradients_ffd_accuracy(self):
        """
        Test :meth:`bet.sensitivity.gradients.calculate_gradients_ffd`.
        """
        self.center_disc = grad.calculate_gradients_ffd(
            self.cluster_disc_ffd, normalize=False)
        self.jacobians = self.center_disc._input_sample_set._jacobians

        nptest.assert_allclose(self.jacobians - self.G_exact, 0,
                               atol=2)

    def test_calculate_gradients_cfd_accuracy(self):
        """
        Test :meth:`bet.sensitivity.gradients.calculate_gradients_cfd`.
        """
        self.center_disc = grad.calculate_gradients_cfd(
            self.cluster_disc_cfd, normalize=False)
        self.jacobians = self.center_disc._input_sample_set._jacobians

        nptest.assert_allclose(self.jacobians - self.G_exact, 0,
                               atol=2)


# Test cases
class test_1to20_1centers_unitsquare(GradientsMethods, unittest.TestCase):

    def setUp(self):
        # Define the input domain (Lambda)
        self.input_dim = 1
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)

        self.lam_domain = np.zeros((self.input_set.get_dim(), 2))
        self.lam_domain[:, 0] = np.zeros(self.input_set.get_dim())
        self.lam_domain[:, 1] = np.ones(self.input_set.get_dim())

        self.input_set.set_domain(self.lam_domain)
        self.input_set_centers.set_domain(self.lam_domain)

        # Choose random centers in input_domian to cluster points around
        self.num_centers = 1
        self.num_close = self.input_set.get_dim() + 1
        self.rvec = 0.1
        np.random.seed(0)
        self.centers = np.random.uniform(self.lam_domain[:, 0],
                                         self.lam_domain[:, 1] - self.lam_domain[:, 0], [self.num_centers,
                                                                                         self.input_set.get_dim()])
        self.input_set_centers.set_values(self.centers)

        # Choose array shapes for RBF methods
        np.random.seed(0)
        self.radii_rbf = np.random.random([self.num_close, self.num_close])
        self.radii_rbfdxi = np.random.random([self.input_dim, self.num_close])
        self.dxi = np.random.random([self.input_dim, self.num_close])

        # Define example linear functions (QoIs) for gradient approximation
        # methods
        self.output_dim = 20
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)


class test_2to20_1centers_unitsquare(GradientsMethods, unittest.TestCase):

    def setUp(self):
        # Define the parameter space (Lambda)
        self.input_dim = 2
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)

        self.lam_domain = np.zeros((self.input_dim, 2))
        self.lam_domain[:, 0] = np.zeros(self.input_dim)
        self.lam_domain[:, 1] = np.ones(self.input_dim)

        self.input_set.set_domain(self.lam_domain)
        self.input_set_centers.set_domain(self.lam_domain)

        # Choose random centers to cluster points around
        self.num_centers = 1
        np.random.seed(0)
        self.centers = np.random.uniform(self.lam_domain[:, 0],
                                         self.lam_domain[:, 1] - self.lam_domain[:, 0], [self.num_centers,
                                                                                         self.input_set.get_dim()])
        self.input_set_centers.set_values(self.centers)
        self.num_close = self.input_dim + 1
        self.rvec = np.random.random(self.input_dim)

        # Choose array shapes for RBF methods
        np.random.seed(0)
        self.radii_rbf = np.random.random([self.num_close, self.num_close])
        self.radii_rbfdxi = np.random.random([self.input_dim, self.num_close])
        self.dxi = np.random.random([self.input_dim, self.num_close])

        # Define example linear functions (QoIs) for gradient approximation
        # methods
        self.output_dim = 20
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)


class test_4to20_100centers_randomhyperbox(
        GradientsMethods, unittest.TestCase):

    def setUp(self):
        # Define the parameter space (Lambda)
        self.input_dim = 4
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)

        self.lam_domain = np.zeros((self.input_dim, 2))
        np.random.seed(0)
        self.lam_domain[:, 0] = np.random.random(self.input_dim)
        self.lam_domain[:, 1] = np.random.random(self.input_dim) + 2

        self.input_set.set_domain(self.lam_domain)
        self.input_set_centers.set_domain(self.lam_domain)

        # Choose random centers to cluster points around
        self.num_centers = 100
        self.centers = np.random.uniform(self.lam_domain[:, 0],
                                         self.lam_domain[:, 1] - self.lam_domain[:, 0], [self.num_centers,
                                                                                         self.input_set.get_dim()])
        self.input_set_centers.set_values(self.centers)
        self.num_close = self.input_set.get_dim() + 1
        self.rvec = 0.1

        # Choose array shapes for RBF methods
        np.random.seed(0)
        self.radii_rbf = np.random.random([self.num_close, self.num_close])
        self.radii_rbfdxi = np.random.random([self.input_dim, self.num_close])
        self.dxi = np.random.random([self.input_dim, self.num_close])

        # Define example linear functions (QoIs) for gradient approximation
        # methods
        self.output_dim = 20
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)


class test_9to20_100centers_randomhyperbox(
        GradientsMethods, unittest.TestCase):

    def setUp(self):
        # Define the parameter space (Lambda)
        self.input_dim = 9
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)

        self.lam_domain = np.zeros((self.input_dim, 2))
        np.random.seed(0)
        self.lam_domain[:, 0] = np.random.random(self.input_dim)
        self.lam_domain[:, 1] = np.random.random(self.input_dim) + 2

        self.input_set.set_domain(self.lam_domain)
        self.input_set_centers.set_domain(self.lam_domain)

        # Choose random centers to cluster points around
        self.num_centers = 100
        self.centers = np.random.uniform(self.lam_domain[:, 0],
                                         self.lam_domain[:, 1] - self.lam_domain[:, 0], [self.num_centers,
                                                                                         self.input_set.get_dim()])
        self.input_set_centers.set_values(self.centers)
        self.num_close = self.input_dim + 1
        self.rvec = 0.1

        # Choose array shapes for RBF methods
        np.random.seed(0)
        self.radii_rbf = np.random.random([self.num_close, self.num_close])
        self.radii_rbfdxi = np.random.random([self.input_dim, self.num_close])
        self.dxi = np.random.random([self.input_dim, self.num_close])

        # Define example linear functions (QoIs) for gradient approximation
        # methods
        self.output_dim = 20
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)


class test_15to37_143centers_negrandomhyperbox(GradientsMethods,
                                               unittest.TestCase):

    def setUp(self):
        # Define the parameter space (Lambda)
        self.input_dim = 15
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)

        self.lam_domain = np.zeros((self.input_dim, 2))
        np.random.seed(0)
        self.lam_domain[:, 0] = -1 * np.random.random(self.input_dim) - 2
        self.lam_domain[:, 1] = -1 * np.random.random(self.input_dim)

        self.input_set.set_domain(self.lam_domain)
        self.input_set_centers.set_domain(self.lam_domain)

        # Choose random centers to cluster points around
        self.num_centers = 143
        self.centers = (self.lam_domain[:, 1] - self.lam_domain[:, 0]) * \
            np.random.random((self.num_centers, self.input_dim)) + \
            self.lam_domain[:, 0]
        self.input_set_centers.set_values(self.centers)
        self.num_close = self.input_dim + 1
        self.rvec = 0.1

        # Choose array shapes for RBF methods
        np.random.seed(0)
        self.radii_rbf = np.random.random([self.num_close, self.num_close])
        self.radii_rbfdxi = np.random.random([self.input_dim, self.num_close])
        self.dxi = np.random.random([self.input_dim, self.num_close])

        # Define example linear functions (QoIs) for gradient approximation
        # methods
        self.output_dim = 37
        coeffs = np.random.random((self.input_dim,
                                   self.output_dim - self.input_dim))
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)


class test_9to30_100centers_randomhyperbox_zeroQoIs(GradientsMethods,
                                                    unittest.TestCase):

    def setUp(self):
        # Define the parameter space (Lambda)
        self.input_dim = 9
        self.input_set = sample.sample_set(self.input_dim)
        self.input_set_centers = sample.sample_set(self.input_dim)

        self.lam_domain = np.zeros((self.input_dim, 2))
        np.random.seed(0)
        self.lam_domain[:, 0] = np.random.random(self.input_dim)
        self.lam_domain[:, 1] = np.random.random(self.input_dim) + 2

        self.input_set.set_domain(self.lam_domain)
        self.input_set_centers.set_domain(self.lam_domain)

        # Choose random centers to cluster points around
        self.num_centers = 100
        self.centers = (self.lam_domain[:, 1] - self.lam_domain[:, 0]) * \
            np.random.random((self.num_centers, self.input_dim)) + \
            self.lam_domain[:, 0]
        self.input_set_centers.set_values(self.centers)
        self.num_close = self.input_dim + 1
        self.rvec = np.random.random(self.input_dim)

        # Choose array shapes for RBF methods
        np.random.seed(0)
        self.radii_rbf = np.random.random([self.num_close, self.num_close])
        self.radii_rbfdxi = np.random.random([self.input_dim, self.num_close])
        self.dxi = np.random.random([self.input_dim, self.num_close])

        # Define example linear functions (QoIs) for gradient approximation
        # methods
        self.output_dim = 30
        coeffs = np.zeros((self.input_dim, 2 * self.input_dim))
        coeffs = np.append(coeffs, np.random.random((self.input_dim,
                                                     self.output_dim - 3 * self.input_dim)), axis=1)
        self.coeffs = np.append(coeffs, np.eye(self.input_dim), axis=1)

# Test cases for the gradient approximation accuracy


class test_2to2_100centers_unitbox(GradientsAccuracy, unittest.TestCase):

    def setUp(self):
        # Define the parameter space (Lambda)
        self.input_dim = 2
        self.input_set_rbf = sample.sample_set(self.input_dim)
        self.input_set_ffd = sample.sample_set(self.input_dim)
        self.input_set_cfd = sample.sample_set(self.input_dim)

        self.input_set_centers = sample.sample_set(self.input_dim)

        self.output_dim = 2
        self.output_set_rbf = sample.sample_set(self.output_dim)
        self.output_set_ffd = sample.sample_set(self.output_dim)
        self.output_set_cfd = sample.sample_set(self.output_dim)

        self.lam_domain = np.zeros((self.input_dim, 2))
        self.lam_domain[:, 0] = np.zeros(self.input_dim)
        self.lam_domain[:, 1] = np.ones(self.input_dim)

        self.input_set_rbf.set_domain(self.lam_domain)
        self.input_set_ffd.set_domain(self.lam_domain)
        self.input_set_cfd.set_domain(self.lam_domain)

        # Choose random centers to cluster points around
        self.num_centers = 100
        np.random.seed(0)
        self.centers = (self.lam_domain[:, 1] - self.lam_domain[:, 0]) * \
            np.random.random((self.num_centers, self.input_dim)) + \
            self.lam_domain[:, 0]
        self.input_set_centers.set_values(self.centers)
        self.num_close = self.input_dim + 1
        self.rvec = 0.01 * np.ones(self.input_dim)

        self.input_set_rbf = grad.sample_l1_ball(self.input_set_centers,
                                                 self.num_close, self.rvec)
        self.input_set_ffd = grad.pick_ffd_points(self.input_set_centers,
                                                  self.rvec)
        self.input_set_cfd = grad.pick_cfd_points(self.input_set_centers,
                                                  self.rvec)

        # Define a vector valued function f : [0,1]x[0,1] -> [x^2, y^2]
        def f(x):
            f = np.zeros(x.shape)
            f[:, 0] = x[:, 0]**2
            f[:, 1] = x[:, 1]**2
            return f

        self.output_set_rbf.set_values(f(self.input_set_rbf.get_values()))
        self.output_set_ffd.set_values(f(self.input_set_ffd.get_values()))
        self.output_set_cfd.set_values(f(self.input_set_cfd.get_values()))
        self.cluster_disc_rbf = sample.discretization(
            self.input_set_rbf, self.output_set_rbf)
        self.cluster_disc_ffd = sample.discretization(
            self.input_set_ffd, self.output_set_ffd)
        self.cluster_disc_cfd = sample.discretization(
            self.input_set_cfd, self.output_set_cfd)

        self.G_exact = np.zeros([self.num_centers, self.output_dim,
                                 self.input_dim])
        self.G_exact[:, 0, 0] = 2 * self.centers[:, 0]
        self.G_exact[:, 1, 1] = 2 * self.centers[:, 1]



#! /usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

"""
This 2D linear example verifies that geometrically distinct QoI can
recreate a probability measure on the input parameter space
used to define the output probability measure.
"""

import numpy as np
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateP as calculateP
import bet.postProcess.plotP as plotP
import bet.postProcess.plotDomains as plotD
import bet.sample as samp
import bet.sampling.basicSampling as bsam
from myModel import my_model

# Define the sampler that will be used to create the discretization
# object, which is the fundamental object used by BET to compute
# solutions to the stochastic inverse problem.
# The sampler and my_model is the interface of BET to the model,
# and it allows BET to create input/output samples of the model.
sampler = bsam.sampler(my_model)

# Initialize 3-dimensional input parameter sample set object
input_samples = samp.sample_set(2)

# Set parameter domain
input_samples.set_domain(np.repeat([[0.0, 1.0]], 2, axis=0))

'''
Suggested changes for user:

Try with and without random sampling.

If using random sampling, try num_samples = 1E3 and 1E4.
What happens when num_samples = 1E2?
Try using 'lhs' instead of 'random' in the 'random_sample_set'.

If using regular sampling, try different numbers of samples
per dimension.
'''
# Generate samples on the parameter space
randomSampling = False
if randomSampling is True:
    input_samples = sampler.random_sample_set(
        'random', input_samples, num_samples=1E3)
else:
    input_samples = sampler.regular_sample_set(
        input_samples, num_samples_per_dim=[30, 30])

'''
Suggested changes for user:

A standard Monte Carlo (MC) assumption is that every Voronoi cell
has the same volume. If a regular grid of samples was used, then
the standard MC assumption is true.

See what happens if the MC assumption is not assumed to be true, and
if different numbers of points are used to estimate the volumes of
the Voronoi cells.
'''
MC_assumption = True
# Estimate volumes of Voronoi cells associated with the parameter samples
if MC_assumption is False:
    input_samples.estimate_volume(n_mc_points=1E5)
else:
    input_samples.estimate_volume_mc()

# Create the discretization object using the input samples
my_discretization = sampler.compute_QoI_and_create_discretization(input_samples,
                                                                  savefile='Validation_discretization.txt.gz')

'''
Compute the output distribution simple function approximation by
propagating a different set of samples to implicitly define a Voronoi
discretization of the data space, corresponding to an implicitly defined
set of contour events defining a discretization of the input parameter
space. The probabilities of the Voronoi cells in the data space (and
thus the probabilities of the corresponding contour events in the
input parameter space) are determined by Monte Carlo sampling using
a set of i.i.d. uniform samples to bin into these cells.

Suggested changes for user:

See the effect of using different values for num_samples_discretize_D.
Choosing

  num_samples_discretize_D = 1

produces exactly the right answer and is equivalent to assigning a
uniform probability to each data sample above (why?).

Try setting this to 2, 5, 10, 50, and 100. Can you explain what you
are seeing? To see an exaggerated effect, try using random sampling
above with n_samples set to 1E2.
'''
num_samples_discretize_D = 1
num_iid_samples = 1E5

Partition_set = samp.sample_set(2)
Monte_Carlo_set = samp.sample_set(2)

Partition_set.set_domain(np.repeat([[0.0, 1.0]], 2, axis=0))
Monte_Carlo_set.set_domain(np.repeat([[0.0, 1.0]], 2, axis=0))

Partition_discretization = sampler.create_random_discretization('random',
                                                                Partition_set,
                                                                num_samples=num_samples_discretize_D)

Monte_Carlo_discretization = sampler.create_random_discretization('random',
                                                                  Monte_Carlo_set,
                                                                  num_samples=num_iid_samples)

# Compute the simple function approximation to the distribution on the
# data space
simpleFunP.user_partition_user_distribution(my_discretization,
                                            Partition_discretization,
                                            Monte_Carlo_discretization)

# Calculate probabilities
calculateP.prob(my_discretization)

########################################
# Post-process the results (optional)
########################################
# Show some plots of the different sample sets
plotD.scatter_2D(my_discretization._input_sample_set,
                 filename='Parameter_Samples',
                 file_extension='.png')
plotD.scatter_2D(my_discretization._output_sample_set,
                 filename='QoI_Samples',
                 file_extension='.png')
plotD.scatter_2D(my_discretization._output_probability_set,
                 filename='Data_Space_Discretization',
                 file_extension='.png')
'''
Suggested changes for user:

At this point, the only thing that should change in the plotP.* inputs
should be either the nbins values or sigma (which influences the kernel
density estimation with smaller values implying a density estimate that
looks more like a histogram and larger values smoothing out the values
more).

There are ways to determine "optimal" smoothing parameters (e.g., see CV, GCV,
and other similar methods), but we have not incorporated these into the code
as lower-dimensional marginal plots generally have limited value in understanding
the structure of a high dimensional non-parametric probability measure.
'''
# calculate 2d marginal probs
(bins, marginals2D) = plotP.calculate_2D_marginal_probs(input_samples,
                                                        nbins=[30, 30])

# plot 2d marginals probs
plotP.plot_2D_marginal_probs(marginals2D, bins, input_samples,
                             filename="validation_raw",
                             file_extension=".png", plot_surface=False)

# smooth 2d marginals probs (optional)
marginals2D = plotP.smooth_marginals_2D(marginals2D, bins, sigma=0.1)

# plot 2d marginals probs
plotP.plot_2D_marginal_probs(marginals2D, bins, input_samples,
                             filename="validation_smooth",
                             file_extension=".png", plot_surface=False)

# calculate 1d marginal probs
(bins, marginals1D) = plotP.calculate_1D_marginal_probs(input_samples,
                                                        nbins=[30, 30])

# plot 1d marginal probs
plotP.plot_1D_marginal_probs(marginals1D, bins, input_samples,
                             filename="validation_raw",
                             file_extension=".png")

# smooth 1d marginal probs (optional)
marginals1D = plotP.smooth_marginals_1D(marginals1D, bins, sigma=0.1)

# plot 1d marginal probs
plotP.plot_1D_marginal_probs(marginals1D, bins, input_samples,
                             filename="validation_smooth",
                             file_extension=".png")



# Copyright (C) 2014-2020 The BET Development Team

"""
This module provides a workaround for people without mpi4py installed
to run BET.
"""

import collections


class comm_for_no_mpi4py(object):

    """
    Provides a fake MPI.COMM_WORLD implementation so that the user need not
    install mpi4py.
    """

    def __init__(self):
        """
        Initialization
        """
        #: size, 1
        self.size = 1
        #: rank, 0
        self.rank = 0
        pass

    def Get_size(self):
        """

        :rtype: int
        :returns: 1

        """
        return 1

    def Get_rank(self):
        """

        :rtype: int
        :returns: 0

        """
        return 0

    def allgather(self, val):
        """
        :param object val: object to allgather

        :rtype: object
        :returns: val

        """
        return [val]

    def gather(self, val1, root=0):
        """
        :param object val1: object to gather
        :param int root: 0

        :rtype: object
        :returns: val1

        """
        return [val1]

    def allreduce(self, val1, op=None):
        """
        :param object val1: object to allreduce

        :rtype: object
        :returns: val1

        """
        return val1

    def bcast(self, val, root=0):
        """
        :param object val: object to broadcast
        :param int root: 0

        :rtype: object
        :returns: val

        """
        return val

    def scatter(self, val1, root=0):
        """
        :param object val1: object to scatter
        :param int root: 0

        :rtype: object
        :returns: val1

        """
        if isinstance(val1, collections.Iterable) and len(val1) == 1:
            val1 = val1[0]
        return val1

    def Allgather(self, val, val2=None):
        """
        :param object val: object to Allgather

        :rtype: object
        :returns: val

        """
        return val

    def Allreduce(self, val1, val2=None, op=None):
        """
        :param object val1: object to Allreduce
        :param object val2: object to Allreduce
        :param op: None

        :rtype: object
        :returns: val1

        """
        return val1

    def Bcast(self, val, root=0):
        """
        :param object val: object to gather
        :param int root: 0

        :rtype: object
        :returns: val

        """
        return val

    def Scatter(self, val1, val2, root=0):
        """
        :param object val1: object to Scatter
        :param object val2: object to Scatter
        :param int root: 0

        :rtype: object
        :returns: val1

        """
        return val1

    def Barrier(self):
        """
        Does nothing in serial.
        """
        pass

    def barrier(self):
        """
        Does nothing in serial.
        """
        pass


class MPI_for_no_mpi4py(object):

    """
    Provides a fake MPI implementation so that the user need not install
    mpi4py.
    """

    def __init__(self):
        """
        Initialization
        """
        #: fake sum
        self.SUM = None
        self.MAX = None
        #: float type
        self.DOUBLE = float
        #: int type
        self.INT = int
        #: bool type
        self.BOOL = bool


try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except ImportError:
    MPI = MPI_for_no_mpi4py()
    comm = comm_for_no_mpi4py()

size = comm.Get_size()
rank = comm.Get_rank()



# Copyright (C) 2014-2020 The BET Development Team

"""
This module provides methods for postprocessing probabilities and data.
"""
import logging
import numpy as np
import bet.sample as sample


class dim_not_matching(Exception):
    """
    Exception for when the dimension is inconsistent.
    """


class bad_object(Exception):
    """
    Exception for when the wrong type of object is used.
    """


def sort_by_rho(sample_set):
    """
    This sorts the samples within the sample_set by probability density.
    If a discretization object is given, then the QoI data is also sorted
    to maintain the correspondence.
    Any volumes present in the input space (or just the sample object)
    are also sorted.

    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base` or
        :class:`~bet.sample.discretization`
    :param indices: sorting indices
    :type indices: :class:`numpy.ndarray` of shape (num_samples,)
    :param sample_set_out: Object containing sorted samples and probabilities
    :type sample_set_out: :class:`~bet.sample.sample_set` or
        :class:`~bet.sample.discretization`

    :rtype: tuple
    :returns: (sample_set_out, indicices)

    """
    if isinstance(sample_set, sample.discretization):
        samples = sample_set._input_sample_set.get_values()
        P_samples = sample_set._input_sample_set.get_probabilities()
        lam_vol = sample_set._input_sample_set.get_volumes()
        data = sample_set._output_sample_set.get_values()
    elif isinstance(sample_set, sample.sample_set_base):
        samples = sample_set.get_values()
        P_samples = sample_set.get_probabilities()
        lam_vol = sample_set.get_volumes()
        data = None
    else:
        raise bad_object("Improper sample object")

    nnz = np.sum(P_samples > 0)
    if lam_vol is None:
        indices = np.argsort(P_samples)[::-1][0:nnz]
    else:
        indices = np.argsort(P_samples / lam_vol)[::-1][0:nnz]
    P_samples = P_samples[indices]
    samples = samples[indices, :]
    if lam_vol is not None:
        lam_vol = lam_vol[indices]
    if data is not None:
        data = data[indices, :]

    if isinstance(sample_set, sample.discretization):
        samples_out = sample.sample_set(sample_set._input_sample_set.get_dim())
        data_out = sample.sample_set(sample_set._output_sample_set.get_dim())
        sample_set_out = sample.discretization(samples_out, data_out)
        sample_set_out._input_sample_set.set_values(samples)
        sample_set_out._input_sample_set.set_probabilities(P_samples)
        sample_set_out._input_sample_set.set_volumes(lam_vol)
        sample_set_out._output_sample_set.set_values(data)
    else:
        sample_set_out = sample.sample_set(sample_set.get_dim())
        sample_set_out.set_values(samples)
        sample_set_out.set_probabilities(P_samples)
        sample_set_out.set_volumes(lam_vol)

    return (sample_set_out, indices)


def sample_prob(percentile, sample_set, sort=True, descending=False):
    """
    This calculates the highest/lowest probability samples whose probability
    sum to a given value.
    A new sample_set with the samples corresponding to these highest/lowest
    probability samples is returned along with the number of samples and
    the indices.
    This uses :meth:`~bet.postProcess.sort_by_rho`.
    The ``descending`` flag determines whether or not to calcuate the
    highest/lowest.

    :param percentile: ratio of highest probability samples to select
    :type percentile: float
    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base` or
        :class:`~bet.sample.discretization`
    :type indices: :class:`numpy.ndarray` of shape (num_samples,)
    :param indices: sorting indices
    :param bool sort: Flag whether or not to sort
    :param bool descending: Flag order of sorting
    :param sample_set_out: Object containing sorted samples and probabilities
    :type sample_set_out: :class:`~bet.sample.sample_set` or
        :class:`~bet.sample.discretization`

    :rtype: tuple
    :returns: ( num_samples, sample_set_out, data)

    """
    if isinstance(sample_set, sample.discretization):
        samples = sample_set._input_sample_set.get_values()
        P_samples = sample_set._input_sample_set.get_probabilities()
        lam_vol = sample_set._input_sample_set.get_volumes()
        data = sample_set._output_sample_set.get_values()
    elif isinstance(sample_set, sample.sample_set_base):
        samples = sample_set.get_values()
        P_samples = sample_set.get_probabilities()
        lam_vol = sample_set.get_volumes()
        data = None
    else:
        raise bad_object("Improper sample object")

    if sort:
        (sample_set, indices) = sort_by_rho(sample_set)
        if isinstance(sample_set, sample.discretization):
            samples = sample_set._input_sample_set.get_values()
            P_samples = sample_set._input_sample_set.get_probabilities()
            lam_vol = sample_set._input_sample_set.get_volumes()
            data = sample_set._output_sample_set.get_values()
        elif isinstance(sample_set, sample.sample_set_base):
            samples = sample_set.get_values()
            P_samples = sample_set.get_probabilities()
            lam_vol = sample_set.get_volumes()
            data = None
    if descending:
        P_samples = P_samples[::-1]
        samples = samples[::-1]
        if lam_vol is not None:
            lam_vol = lam_vol[::-1]
        if data is not None:
            data = data[::-1]
        indices = indices[::-1]

    P_sum = np.cumsum(P_samples)
    num_samples = np.sum(np.logical_and(0.0 < P_sum, P_sum <= percentile))
    P_samples = P_samples[0:num_samples]
    samples = samples[0:num_samples, :]
    if lam_vol is not None:
        lam_vol = lam_vol[0:num_samples]
    if data is not None:
        if len(data.shape) == 1:
            data = np.expand_dims(data, axis=1)
        data = data[0:num_samples, :]

    if isinstance(sample_set, sample.discretization):
        samples_out = sample.sample_set(sample_set._input_sample_set.get_dim())
        data_out = sample.sample_set(sample_set._output_sample_set.get_dim())
        sample_set_out = sample.discretization(samples_out, data_out)
        sample_set_out._input_sample_set.set_values(samples)
        sample_set_out._input_sample_set.set_probabilities(P_samples)
        sample_set_out._input_sample_set.set_volumes(lam_vol)
        sample_set_out._output_sample_set.set_values(data)
    else:
        sample_set_out = sample.sample_set(sample_set.get_dim())
        sample_set_out.set_values(samples)
        sample_set_out.set_probabilities(P_samples)
        sample_set_out.set_volumes(lam_vol)

    return (num_samples, sample_set_out,
            indices[0:num_samples])


def sample_highest_prob(top_percentile, sample_set, sort=True):
    """
    This calculates the highest probability samples whose probability sum to a
    given value.
    The number of high probability samples that sum to the value,
    a new sample_set, and the indices are returned.
    This uses :meth:`~bet.postProcess.sort_by_rho`.

    :param top_percentile: ratio of highest probability samples to select
    :type top_percentile: float
    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :type indices: :class:`numpy.ndarray` of shape (num_samples,)
    :param indices: sorting indices
    :param bool sort: Flag whether or not to sort
    :param sample_set_out: Object containing sorted samples and probabilities
    :type sample_set_out: :class:`~bet.sample.sample_set`
        or :class:`~bet.sample.discretization`

    :rtype: tuple
    :returns: ( num_samples, sample_set_out, indices)

    """
    return sample_prob(top_percentile, sample_set, sort)


def sample_lowest_prob(bottom_percentile, sample_set, sort=True):
    """
    This calculates the lowest probability samples whose probability sum to a
    given value.
    The number of low probability samples that sum to the value,
    a new sample_set, and the indices are returned.
    This uses :meth:`~bet.postProcess.sort_by_rho`.

    :param top_percentile: ratio of highest probability samples to select
    :type top_percentile: float
    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :type indices: :class:`numpy.ndarray` of shape (num_samples,)
    :param indices: sorting indices of unsorted ``P_samples``
    :param bool sort: Flag whether or not to sort
    :param sample_set_out: Object containing sorted samples and probabilities
    :type sample_set_out: :class:`~bet.sample.sample_set`
        or :class:`~bet.sample.discretization`

    :rtype: tuple
    :returns: ( num_samples, sample_set_out, indices)

    """
    return sample_prob(bottom_percentile, sample_set,
                       sort, descending=True)




# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains tests for :module:`bet.postProcess.postTools`.


Tests for correct post-processing.
"""
import unittest
import bet.calculateP.simpleFunP as simpleFunP
import bet.postProcess.postTools as postTools
import numpy as np
import scipy.spatial as spatial
import numpy.testing as nptest
import bet.util as util
from bet.Comm import comm
import bet.sample as sample

'''
def test_in_high_prob():
    """

    Tests :meth:`bet.postProcess.postTools.in_high_prob`
    """
    def rho_D(my_data):
        return my_data/4.0
    output_samples = sample.sample_set(1)\
    output_samples.set_values(np.array([0, 1, 0, 1, 1, 1]))
    maximum = np.max(rho_D(output_samples.get_values()))
    print "maximum", maximum
    assert 4 == postTools.in_high_prob(output_samples.get_values(), rho_D, maximum)
    assert 3 == postTools.in_high_prob(output_samples.get_values(), rho_D, maximum, [3, 4, 5])
    assert 2 == postTools.in_high_prob(output_samples.get_values(), rho_D, maximum, [0, 1, 2, 3])
    assert 1 == postTools.in_high_prob(output_samples.get_values(), rho_D, maximum, [0, 2, 4])
    assert 0 == postTools.in_high_prob(output_samples.get_values(), rho_D, maximum, [0, 2])

def test_in_high_prob_multi():
    """

    Tests :meth:`bet.postProcess.postTools.in_high_prob_multi`

    """
    def rho_D(my_data):
        return my_data/4.0
    data1 = np.array([0, 1, 0, 1, 1, 0])
    data2 = np.ones(data1.shape)-data1
    maximum = np.max(rho_D(data1))

    print "maximum", maximum
    results_list = [[None, data1], [None, data2], [None, data1], [None, data2]]
    sample_nos_list = [[3, 4, 5], [3, 4, 5], [0, 2, 4], [0, 2, 4]]

    nptest.assert_array_equal(np.array([2, 1, 1, 2]),
            postTools.in_high_prob_multi(results_list, rho_D, maximum,
                sample_nos_list))
    nptest.assert_array_equal(np.array([3, 3, 3, 3]),
            postTools.in_high_prob_multi(results_list, rho_D, maximum))

def test_compare_yield_CH():
    """

    Tests :meth:`bet.postProcess.postTools.compare_yield` with column headings

    """
    sample_quality = np.random.random((10,))
    sort_ind = np.argsort(sample_quality)
    run_param = []
    for i in range(10):
        run_param.append(np.random.random((4,)))
    column_headings = ['swallow', 'coconut', 'ni', 'shrubbery']
    try:
        postTools.compare_yield(sort_ind, sample_quality, run_param,
                column_headings)
        go = True
    except (RuntimeError, TypeError, NameError):
        go = False
    nptest.assert_equal(go, True)

def test_compare_yield():
    """

    Tests :meth:`bet.postProcess.postTools.compare_yield` without column headings

    """
    sample_quality = np.random.random((10,))
    sort_ind = np.argsort(sample_quality)
    run_param = []
    for i in range(10):
        run_param.append(np.random.random((4,)))
    try:
        postTools.compare_yield(sort_ind, sample_quality, run_param)
        go = True
    except (RuntimeError, TypeError, NameError):
        go = False
    nptest.assert_equal(go, True)
'''


class Test_PostTools(unittest.TestCase):
    """
    Test :mod:`bet.postProcess.postTools`.
    """

    def setUp(self):
        """
        Set up problem.
        """
        input_samples = sample.sample_set(1)
        input_samples.set_domain(np.array([[0.0, 1.0]]))
        # self.lam_domain=np.array([[0.0,1.0]])
        num_samples = 1000
        input_samples.set_values(np.linspace(input_samples.get_domain()[0, 0],
                                             input_samples.get_domain()[0, 1],
                                             num_samples + 1))
        #self.samples = np.linspace(self.lam_domain[0][0], self.lam_domain[0][1], num_samples+1)
        input_samples.set_probabilities((1.0 / float(input_samples.get_values().shape[0])) *
                                        np.ones((input_samples.get_values().shape[0],)))
        #self.P_samples = (1.0/float(self.samples.shape[0]))*np.ones((self.samples.shape[0],))
        input_samples._probabilities[0] = 0.0
        input_samples._probabilities[-1] *= 2.0
        #self.P_samples[0] = 0.0
        #self.P_samples[-1] *= 2.0

        self.data = input_samples
        #self.data = self.samples[:]

    def test_sort_by_rho(self):
        """
        Test :meth:`bet.postProcess.postTools.sort_by_rho`.
        """
        disc = sample.discretization(input_sample_set=self.data,
				     output_sample_set=self.data.copy())
        (self.data1, _) = postTools.sort_by_rho(disc)
        self.assertGreater(np.min(self.data1.get_input_sample_set().
                                  get_probabilities()), 0.0)
        nptest.assert_almost_equal(np.sum(self.data1.get_input_sample_set().
                                   get_probabilities()), 1.0)

        (self.data2, _) = postTools.sort_by_rho(self.data)
        self.assertGreater(np.min(self.data2.get_probabilities()), 0.0)
        nptest.assert_almost_equal(np.sum(self.data2.get_probabilities()), 1.0)

    def test_sample_prob(self):
        """
        Test :meth:`bet.postProcess.postTools.sample_prob`.
        """
        (num_samples, sample_set_out, _) = postTools.sample_prob(1.0, self.data,
                                                                 sort=True,
                                                                 descending=True)

        nptest.assert_almost_equal(
            np.sum(sample_set_out.get_probabilities()), 1.0)
        nptest.assert_equal(num_samples, 1000)

        (num_samples, sample_set_out, _) = postTools.sample_prob(0.8,
                                                                 self.data,
                                                                 sort=True,
                                                                 descending=True)

        nptest.assert_allclose(
            np.sum(sample_set_out.get_probabilities()), 0.8, 0.001)

        (num_samples, sample_set_out, _) = postTools.sample_prob(1.0,
                                                                 self.data,
                                                                 sort=True,
                                                                 descending=False)

        nptest.assert_almost_equal(
            np.sum(sample_set_out.get_probabilities()), 1.0)
        nptest.assert_equal(num_samples, 1000)

        (num_samples, sample_set_out, _) = postTools.sample_prob(0.8,
                                                                 self.data,
                                                                 sort=True,
                                                                 descending=False)

        nptest.assert_allclose(
            np.sum(sample_set_out.get_probabilities()), 0.8, 0.001)

    def test_sample_highest_prob(self):
        """
        Test :meth:`bet.postProcess.postTools.sample_highest_prob`.
        """
        (num_samples, sample_set_out, _) = postTools.sample_highest_prob(1.0,
                                                                         self.data,
                                                                         sort=True)

        nptest.assert_almost_equal(
            np.sum(sample_set_out.get_probabilities()), 1.0)
        nptest.assert_equal(num_samples, 1000)

        (num_samples, sample_set_out, _) = postTools.sample_highest_prob(0.8,
                                                                         self.data,
                                                                         sort=True)

        nptest.assert_allclose(
            np.sum(sample_set_out.get_probabilities()), 0.8, 0.001)

    def test_sample_lowest_prob(self):
        """
        Test :meth:`bet.postProcess.postTools.sample_lowest_prob`.
        """
        (num_samples, sample_set_out, _) = postTools.sample_lowest_prob(1.0,
                                                                        self.data,
                                                                        sort=True)

        nptest.assert_almost_equal(
            np.sum(sample_set_out.get_probabilities()), 1.0)
        nptest.assert_equal(num_samples, 1000)

        (num_samples, sample_set_out, _) = postTools.sample_lowest_prob(0.8,
                                                                        self.data,
                                                                        sort=True)

        nptest.assert_allclose(
            np.sum(sample_set_out.get_probabilities()), 0.8, 0.001)



# Copyright (C) 2014-2020 The BET Development Team

"""
This subpackage contains the test modules for the sampling subpackage.
"""
__all__ = ['test_basicSampling',
           'test_Lp_generalized_samples', 'test_useLUQ']
